"""Sample Tool for Agent Builder."""
import logging

import requests
from firebase_admin import initialize_app
from firebase_functions import https_fn
from flask import Flask, request
from google.auth import default as google_default_auth
from google.auth.transport.requests import Request as GoogleRequest

initialize_app()
app = Flask(__name__)

# Set up logging
logging.basicConfig(level=logging.INFO)

# Google Auth
creds, project = google_default_auth()
auth_req = GoogleRequest()

def get_request_params(param, default=None):
    request_json = request.get_json(silent=True)
    request_args = request.args

    return (request_json or {}).get(param, default) or request_args.get(
        param, default)

@app.route("/get_weather_grid", methods=["GET", "POST"])
def get_weather_grid():
    lat = get_request_params("latitude")
    long = get_request_params("longitude")

    url = f"https://api.weather.gov/points/{lat},{long}"
    response = requests.get(url)
    logging.info(f"FIRST RESPONSE: {response}")
    if response.status_code == 200:
        data = response.json()
        logging.info(f"FIRST RESPONSE DATA: {data}")

        response = requests.get(data["properties"]["forecast"])
        logging.info(f"SECOND RESPONSE: {response}")
        if response.status_code == 200:
            data = response.json()
            logging.info(f"SECOND RESPONSE DATA: {data}")

            # just take the weather from the current period
            period = data["properties"]["periods"][0]
            return {
                "temperature": period["temperature"],
                "temperatureUnit": period["temperatureUnit"]
                }

    return {}

@https_fn.on_request()
def main(req: https_fn.Request) -> https_fn.Response:
    creds.refresh(auth_req)
    with app.request_context(req.environ):
        return app.full_dispatch_request()



"""Evaluation tooling for Vertex Conversation DataStores native service."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import io
import itertools
import json
import math
import os
import re
from datetime import datetime, timezone
from typing import Union

import matplotlib.pyplot as plt
import pandas as pd
import plotly.graph_objects as go
from google.cloud import bigquery
from googleapiclient.discovery import build
from googleapiclient.http import MediaInMemoryUpload, MediaIoBaseDownload
from tqdm import tqdm

from dfcx_scrapi.core.scrapi_base import ScrapiBase
from dfcx_scrapi.tools.agent_response import AgentResponse
from dfcx_scrapi.tools.metrics import build_metrics

_FOLDER_ID = re.compile(r"folders\/(.*?)(?=\/|\?|$)")
EVAL_RESULTS_COLS = [
            "answer_generator_llm_rendered_prompt",
            "search_results"
            ]

class DataStoreEvaluator(ScrapiBase):
    def __init__(
        self,
        metrics: list[str],
        project_id: str,
        location_id: str,
        model: str = "gemini-2.0-flash-001"
        ):
        self.project_id = project_id
        self.location_id = location_id
        self.metrics = build_metrics(
            metrics=metrics,
            model_id= model,
            genai_client = ScrapiBase._get_genai_client(
                self.project_id,
                self.location_id
                )
        )

    def run(self, scraper_output: pd.DataFrame) -> "EvaluationResult":
        timestamp = datetime.now(tz=timezone.utc)
        scraper_output = scraper_output.copy(deep=True)
        result = pd.DataFrame(index=scraper_output.index)

        for metric in self.metrics:
           result = pd.concat([result, metric.run(scraper_output)], axis=1)

        # adding timestamp and agent display name so they can be used as a multi
        # index
        result["evaluation_timestamp"] = timestamp.isoformat()

        return EvaluationResult(scraper_output, result)
@dataclasses.dataclass
class EvaluationResult:
    scrape_outputs: pd.DataFrame = None
    metric_outputs: pd.DataFrame = None

    @property
    def timestamp(self) -> str:
        return self.metric_outputs["evaluation_timestamp"].iloc[0]

    @staticmethod
    def truncate(df, column):
        truncated_fix = "<TRUNCATED: Google Sheet 50k character limit>"
        def _truncate(value):
            if len(value) < 50_000:
                return value
            else:
                return value[:50_000 - len(truncated_fix)] + truncated_fix
        df[column] = df[column].apply(_truncate)

    @staticmethod
    def find_folder(folder_name, drive_service) -> Union[tuple[str, str], None]:
        """Finds a folder by name in Google Drive."""
        query = (
            f"name = '{folder_name}' and "
            f"mimeType = 'application/vnd.google-apps.folder' and "
            f"trashed = false"
        )
        fields = "nextPageToken, files(id, name, webViewLink)"
        list_request = drive_service.files().list(q=query, fields=fields)
        result = list_request.execute()
        folders = result.get("files", [])
        if not folders:
            return None

        return folders[0].get("id"), folders[0].get("webViewLink")

    @staticmethod
    def create_folder(
        folder_name, drive_service
        ) -> tuple[Union[str, None], Union[str, None]]:
        """Creates a folder in Google Drive."""
        create_request = drive_service.files().create(
            body={
                "name": folder_name,
                "mimeType": "application/vnd.google-apps.folder"
            },
            fields="id, webViewLink"
        )
        result = create_request.execute()

        return result.get("id"), result.get("webViewLink")

    @staticmethod
    def create_json(
            content, file_name, parent, drive_service
            ) -> tuple[Union[str, None], Union[str, None]]:
        """Creates a .json file in the specified Google Drive folder."""
        request = drive_service.files().create(
            body={"name": file_name, "parents": [parent]},
            media_body=MediaInMemoryUpload(
                json.dumps(content, indent=4).encode("utf-8"),
                mimetype="text/plain",
            ),
            fields="id, webViewLink",
        )
        result = request.execute()

        return result.get("id"), result.get("webViewLink")

    @staticmethod
    def create_chunks(iterable, chunk_size):
        for chunk in itertools.zip_longest(*([iter(iterable)] * chunk_size)):
            yield [element for element in chunk if element is not None]

    @staticmethod
    def delete_worksheet(sheet_id, worksheet_id, sheets_service):
        """Deletes a worksheet."""
        sheets_service.spreadsheets().batchUpdate(
            spreadsheetId=sheet_id,
            body={"requests": [{"deleteSheet": {"sheetId": worksheet_id}}]},
        ).execute()

    @staticmethod
    def get_bigquery_types(df):
        "Maps DataFrame data types to BigQuery data types."
        types = []
        data_type_mapping = {
            'object': 'STRING',
            'int64': 'INTEGER',
            'float64': 'FLOAT',
            'bool': 'BOOLEAN',
            'datetime64[ns]': 'TIMESTAMP'  # Assuming nanosecond timestamps
            }
        for dtype in df.dtypes:
            if dtype in data_type_mapping:
                types.append(data_type_mapping[dtype])
            else:
                # Handle other data types (error handling or placeholder)
                types.append('STRING')  # Placeholder, adjust as needed
                print(f"Warning: Unhandled data type: {dtype}")

        return types

    @staticmethod
    def sanitize_column_names(df):
        "Sanitizes column names replacing special characters with underscores."
        sanitized_names = []
        for col in df.columns:
            # Replace special characters with underscores
            sanitized_name = re.sub(r"[^\w\s]", "_", col)
            sanitized_names.append(sanitized_name)

        return df.rename(columns=dict(zip(df.columns, sanitized_names)))

    @staticmethod
    def list_folder(folder_id, drive_service) -> list[tuple[str, str]]:
        query = f"'{folder_id}' in parents and trashed = false"
        list_request = drive_service.files().list(
            q=query, fields="nextPageToken, files(id, name)"
        )
        result = list_request.execute()
        items = result.get("files", [])
        return [(item["id"], item["name"]) for item in items]

    @staticmethod
    def download_json(file_id, drive_service):
        request = drive_service.files().get_media(fileId=file_id)
        fh = io.BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while not done:
            _, done = downloader.next_chunk()

        fh.seek(0)

        return json.loads(fh.read().decode('utf-8'))

    def load(self, folder_url, credentials):
        folder_id_match = _FOLDER_ID.search(folder_url)
        if not folder_id_match:
            raise ValueError()

        folder_id = folder_id_match.group(1)
        drive_service = build("drive", "v3", credentials=credentials)

        file_id = self.find_file_in_folder(
            folder_id, "results.json", drive_service)
        json_content = self.download_json(file_id, drive_service)

        queryset = pd.DataFrame.from_dict(
            json_content["queryset"], orient="index")
        responses = pd.DataFrame.from_dict(
            json_content["responses"], orient="index"
        )

        ar = AgentResponse()
        queryset["query_result"] = responses.apply(
            ar.from_row, axis=1
        )
        self.scrape_outputs = queryset

        self.metric_outputs = pd.DataFrame.from_dict(
            json_content["metrics"], orient="index"
        )

    def aggregate(self, columns: list[str] = None):
        if not columns:
            columns = self.metric_outputs.columns
        shared_columns = self.metric_outputs.columns.intersection(set(columns))
        result = pd.DataFrame(self.metric_outputs[shared_columns])
        result["name"] = self.scrape_outputs["agent_display_name"]
        result["evaluation_timestamp"] = (
            self.metric_outputs["evaluation_timestamp"]
        )
        result = result.set_index(["name", "evaluation_timestamp"])

        return result.groupby(level=[0, 1]).mean(numeric_only=True)

    def export(self, folder_name: str, chunk_size: int, credentials):
        drive_service = build("drive", "v3", credentials=credentials)
        folder = self.find_folder(folder_name, drive_service)
        if folder:
            folder_id, folder_url = folder
        else:
            folder_id, folder_url = self.create_folder(
                folder_name, drive_service
                )

        queryset = self.scrape_outputs.drop("query_result", axis=1)
        responses = self.scrape_outputs["query_result"].apply(
            lambda x: x.to_row()
            )
        responses = pd.DataFrame(responses.to_list(), index=queryset.index)

        json_content = {
            "queryset": queryset.to_dict(orient="index"),
            "responses": responses.to_dict(orient="index"),
            "metrics": self.metric_outputs.to_dict(orient="index"),
        }
        json_id, json_url = self.create_json(
            json_content, "results.json", folder_id, drive_service
        )

        for column in EVAL_RESULTS_COLS:
            self.truncate(responses, column)

        results = pd.concat([queryset, responses, self.metric_outputs], axis=1)
        worksheets = {
            "summary": self.aggregate().fillna("#N/A"),
            "results": results.fillna("#N/A")
        }
        sheets_service = build("sheets", "v4", credentials=credentials)
        self.create_sheet(
            worksheets=worksheets,
            title="results",
            parent=folder_id,
            chunk_size=chunk_size,
            sheets_service=sheets_service,
            drive_service=drive_service,
        )
        return folder_url

    def export_to_csv(self, file_name: str):
        queryset = self.scrape_outputs.drop("query_result", axis=1)
        responses = self.scrape_outputs["query_result"].apply(
            lambda x: x.to_row())
        responses = pd.DataFrame(responses.to_list(), index=queryset.index)

        for column in EVAL_RESULTS_COLS:
            self.truncate(responses, column)

        results = pd.concat([queryset, responses, self.metric_outputs], axis=1)
        temp_dir = "/tmp/evaluation_results"
        os.makedirs(temp_dir, exist_ok=True)
        filepath = os.path.join(temp_dir, file_name)
        results.to_csv(filepath, index=False)

        return filepath

    def display_on_screen(self):
        queryset = self.scrape_outputs.drop("query_result", axis=1)
        responses = self.scrape_outputs["query_result"].apply(
            lambda x: x.to_row())
        responses = pd.DataFrame(responses.to_list(), index=queryset.index)

        for column in EVAL_RESULTS_COLS:
            self.truncate(responses, column)

        results = pd.concat([queryset, responses, self.metric_outputs], axis=1)

        return results

    def export_to_bigquery(
            self,
            eval_results,
            project_id: str,
            dataset_id: str,
            table_name: str,
            credentials
            ):
        data=eval_results.scrape_outputs["query_result"].apply(
            lambda x: x.to_row())
        data = pd.DataFrame(data.to_list(),eval_results.scrape_outputs.index)
        eval_results.scrape_outputs["query_result"] = None
        df = pd.concat(
            [
                data,
                eval_results.scrape_outputs,
                eval_results.metric_outputs
                ],
                axis=1)

        df = EvaluationResult.sanitize_column_names(df)
        client = bigquery.Client(project=project_id, credentials=credentials)

        try:
            df['conversation_id'] = df['conversation_id'].astype(str)
            df['latency'] = df['latency'].astype(str)
            df['expected_uri'] = df['expected_uri'].astype(str)
            df['answerable'] = df['answerable'].astype(str)
            df['golden_snippet'] = df['golden_snippet'].astype(str)

            df = df.drop('query_result', axis=1)
            df = df.drop('golden_snippet', axis=1)
            df = df.drop('answerable', axis=1)

            load_job = client.load_table_from_dataframe(df, '.'.join(
                [project_id, dataset_id, table_name]))

            return load_job.result()
        except Exception as e:
            print(f"Error exporting data: {e}")
            return None  # Indicate failure

    def find_file_in_folder(
            self,
            folder_id,
            name,
            drive_service
            ) -> Union[str, None]:
        for file_id, file_name in self.list_folder(folder_id, drive_service):
            if file_name == name:
                return file_id
        return None

    def add_worksheet(
            self, sheet_id, content, title, sheets_service, chunk_size) -> None:
        """Adds a worksheet to an existing spreadsheet."""
        sheets_service.spreadsheets().batchUpdate(
            spreadsheetId=sheet_id,
            body={"requests": [{"addSheet": {"properties": {"title": title}}}]},
        ).execute()

        for chunk in tqdm(
            self.create_chunks(content, chunk_size),
            total=math.ceil(len(content) / chunk_size),
            desc=f"Creating worksheet: {title}",
            ):
            sheets_service.spreadsheets().values().append(
                spreadsheetId=sheet_id,
                range=f"'{title}'!A1",
                valueInputOption="RAW",
                body={"values": chunk},
            ).execute()

    def create_sheet(
            self, worksheets, title, parent, chunk_size, sheets_service,
            drive_service) -> Union[str, None]:
        """Creates a new spreadsheet with worksheets."""
        body = {"properties": {"title": title}}
        create_request = sheets_service.spreadsheets().create(
            body=body, fields="spreadsheetId"
        )
        create_result = create_request.execute()
        sheet_id = create_result.get("spreadsheetId")

        parents_request = drive_service.files().get(
            fileId=sheet_id, fields="parents")
        parents_result = parents_request.execute()
        parents = parents_result.get("parents")
        previous_parents = ",".join(parents) if parents else None

        if not sheet_id:
            return

        for worksheet_title, content in worksheets.items():
            content_dict = content.to_dict(orient="split")
            self.add_worksheet(
                sheet_id=sheet_id,
                content=[content_dict["columns"]] + content_dict["data"],
                title=worksheet_title,
                sheets_service=sheets_service,
                chunk_size=chunk_size,
            )

        all_request = sheets_service.spreadsheets().get(spreadsheetId=sheet_id)
        all_result = all_request.execute()
        default_sheet_id = all_result["sheets"][0]["properties"]["sheetId"]

        self.delete_worksheet(sheet_id, default_sheet_id, sheets_service)
        _ = drive_service.files().update(
            fileId=sheet_id,
            addParents=parent,
            removeParents=previous_parents,
            fields="id, parents"
            ).execute()

        return f"https://docs.google.com/spreadsheets/d/{sheet_id}/edit"


@dataclasses.dataclass
class EvaluationVisualizer:
  evaluation_results: list[EvaluationResult]

  def radar_plot(self, columns: Union[list[str], None] = None):
    fig = go.Figure()
    summaries = pd.concat(
        [result.aggregate(columns) for result in self.evaluation_results]
    )
    summaries = summaries.to_dict(orient="split")

    for idx, values in enumerate(summaries["data"]):
      fig.add_trace(
          go.Scatterpolar(
              r=values,
              theta=summaries["columns"],
              fill='toself',
              name="_".join(summaries["index"][idx]),
          )
      )
    fig.update_layout(
        polar={"radialaxis": {"visible": True, "range": [0, 1]}},
        showlegend=True
    )
    fig.show()

  def count_barplot(self, column_name: str):
    results = []
    for result in self.evaluation_results:
      responses = result.scrape_outputs["query_result"].apply(
          lambda x: x.to_row())
      responses = pd.DataFrame(
          responses.to_list(), index=result.scrape_outputs.index
      )
      results.append(
          pd.concat(
              [result.scrape_outputs, responses, result.metric_outputs],
              axis=1
          )
      )
    results = pd.concat(results)
    results = results.set_index(["agent_display_name", "evaluation_timestamp"])
    grouped_counts = (
        results[column_name]
        .groupby(level=["agent_display_name", "evaluation_timestamp"])
        .value_counts()
        .unstack(fill_value=0)
    )
    grouped_counts.plot(kind="bar")
    plt.xlabel("Name")
    plt.ylabel("Count")
    plt.xticks(rotation=15)
    plt.title(f"{column_name} counts by name")
    plt.legend(title=column_name)
    plt.show()

  def mean_barplot(self, column_names: list[str]):
    results = []
    for result in self.evaluation_results:
      results.append(
          pd.concat([result.scrape_outputs, result.metric_outputs], axis=1)
      )
    results = pd.concat(results)
    results = results.set_index(["agent_display_name", "evaluation_timestamp"])
    grouped_means = (
        results[column_names]
        .groupby(level=["agent_display_name", "evaluation_timestamp"])
        .mean()
    )
    grouped_means.plot(kind="bar")
    plt.ylim(top=1.0)
    plt.xlabel("Name")
    plt.ylabel("Mean")
    plt.xticks(rotation=15)
    plt.title("mean by name")
    plt.show()



"""Unit Tests for Sessions."""
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch
import uuid

import pytest
from google.protobuf import struct_pb2
from google.cloud.dialogflowcx_v3beta1 import types

from dfcx_scrapi.core.sessions import Sessions

@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    location_id = "global"
    agent_id = f"projects/{project_id}/locations/{location_id}/agents/fcdecc6a-3f2e-4f8d-abca-63426024d8bb"
    intent_id = f"{agent_id}/intents/b1983e56-5c96-4b20-b15a-fb5c12b77500"
    flow_id = f"{agent_id}/flows/925c0042-686f-4422-bf34-2a2a37b1a3ee"
    tool_id = f"{agent_id}/tools/8e1205e8-bb5e-46ef-896c-1e4f72bb871c"
    page_id = f"{flow_id}/pages/8e1205e8-bb5e-46ef-896c-1e4f72bb871c"
    playbook_id = f"{agent_id}/playbooks/e79502db-78e0-4f54-8447-716107bb553e"
    playbook_name = "mock_playbook"
    environment_id = f"{agent_id}/environments/753de31b-a14b-45f3-a731-b40831ecfbc4"
    environment_name = "My Test Environment"
    session_id_plain = f"{agent_id}/sessions/a1b2c3d4-e5f6-7890-1234-567890abcdef"
    session_id_with_env = f"{environment_id}/sessions/a1b2c3d4-e5f6-7890-1234-567890abcdef"
    language_code = "en"
    text = "Hello!"
    tool_action = "my_tool_action"
    tool_input_params = {"param1": "value1", "param2": "value2"}
    tool_output_params = {"param3": "value3", "param4": "value4"}
    tool_input_params_nested = {"top_key": {"param1": "value1", "param2": "value2"}}
    tool_output_params_nested = {"top_key": {"param3": "value3", "param4": "value4"}}
    
    parameters = {"param1": "value1", "param2": "value2"}
    end_user_metadata = {"user_id": "user123"}
    
    return {
        "project_id": project_id,
        "location_id": location_id,
        "agent_id": agent_id,
        "session_id_plain": session_id_plain,
        "session_id_with_env": session_id_with_env,
        "intent_id": intent_id,
        "flow_id": flow_id,
        "page_id": page_id,
        "tool_id": tool_id,
        "playbook_id": playbook_id,
        "playbook_name": playbook_name,
        "environment_id": environment_id,
        "environment_name": environment_name,
        "language_code": language_code,
        "text": text,
        "tool_action": tool_action,
        "tool_input_params": tool_input_params,
        "tool_output_params": tool_output_params,
        "tool_input_params_nested": tool_input_params_nested,
        "tool_output_params_nested": tool_output_params_nested,
        "parameters": parameters,
        "end_user_metadata": end_user_metadata,
    }

@pytest.fixture
def mock_query_result(test_config):
    """Create a mock QueryResult object for testing, without generative info."""
    return types.session.QueryResult(
        text=test_config["text"],
        language_code=test_config["language_code"],
        parameters=struct_pb2.Struct(
            fields={
                "some_key": struct_pb2.Value(string_value="some_value")
            }
        ),
        response_messages=[
            types.ResponseMessage(
                text=types.ResponseMessage.Text(
                    text=["Greetings! How can I assist?"]
                )
            ),
              types.ResponseMessage(
                text=types.ResponseMessage.Text(
                    text=["Hey! What can I help you with today?"]
                )
            )
        ],
        intent_detection_confidence=1,
        match=types.session.Match(
            match_type=types.session.Match.MatchType.PLAYBOOK,
            confidence=1
        ),
        advanced_settings = types.AdvancedSettings(
           speech_settings = types.AdvancedSettings.SpeechSettings(
            endpointer_sensitivity=90,
           ),
           logging_settings = types.AdvancedSettings.LoggingSettings(
               enable_stackdriver_logging=True,
                enable_interaction_logging=True
            )
        ),
    )

@pytest.fixture
def mock_query_result_tools_playbooks_flows(test_config):
    """Create a mock QueryResult object for testing, with generative info."""
    return types.session.QueryResult(
        text=test_config["text"],
        language_code=test_config["language_code"],
        generative_info = types.session.GenerativeInfo(
            action_tracing_info=types.Example(
                actions=[
                    types.Action(
                        tool_use=types.trace.ToolUse(
                            tool=test_config["tool_id"],
                            action=test_config["tool_action"],
                            input_action_parameters=test_config["tool_input_params"],
                            output_action_parameters=test_config["tool_output_params"],
                        ),
                    ),
                    types.Action(
                        playbook_invocation=types.trace.PlaybookInvocation(
                            playbook=test_config["playbook_id"]
                        ),
                    ),
                    types.Action(
                        flow_invocation=types.trace.FlowInvocation(
                            flow=test_config["flow_id"]
                        )
                    ),
                    types.Action(
                        agent_utterance=types.AgentUtterance(text="Hey there!")
                    )
                ]
            ),
        current_playbooks=[test_config["playbook_id"]]
        ),
    )

@pytest.fixture
def mock_query_result_datastore(test_config):
    """Create a mock QueryResult object for testing datastore responses."""
    return types.session.QueryResult(
        text="who is the ceo?",
        language_code="en",
        parameters=struct_pb2.Struct(
            fields={
                "some_key": struct_pb2.Value(string_value="some_value")
            }
        ),
        response_messages=[
            types.ResponseMessage(
                text=types.ResponseMessage.Text(
                    text=["Sundar Pichai is the CEO of Google.\nhttps://www.google.com"]
                )
            ),
            types.ResponseMessage(
                payload=struct_pb2.Struct(
                  fields={
                    "richContent": struct_pb2.Value(
                      list_value=struct_pb2.ListValue(
                        values=[
                            struct_pb2.Value(
                            list_value=struct_pb2.ListValue(
                              values=[
                                  struct_pb2.Value(
                                    struct_value=struct_pb2.Struct(
                                      fields={
                                          "type": struct_pb2.Value(string_value="info"),
                                          "title": struct_pb2.Value(string_value="CEO of Google"),
                                          "subtitle": struct_pb2.Value(string_value="Information on Google executive team."),
                                          "metadata": struct_pb2.Value(struct_value=struct_pb2.Struct()),
                                          "actionLink": struct_pb2.Value(string_value="https://www.google.com"),
                                        }
                                    )
                                  )
                                ]
                            )
                          )
                        ]
                      )
                   )
                  }
                )
            )
          ],
        intent_detection_confidence=1,
        match=types.session.Match(
            match_type=types.session.Match.MatchType.KNOWLEDGE_CONNECTOR,
            confidence=1
        ),
        advanced_settings = types.AdvancedSettings(
           logging_settings = types.AdvancedSettings.LoggingSettings(
               enable_stackdriver_logging=True,
                enable_interaction_logging=True
            )
        ),
        current_page = types.Page(
          name=test_config["page_id"],
          display_name="Start Page"
        )
    )

@pytest.fixture
def mock_query_result_params_no_text_input(test_config):
    """Create a mock QueryResult object for testing."""
    return types.session.QueryResult(
        language_code=test_config["language_code"],
    )

@pytest.fixture
def mock_detect_intent_response(mock_query_result):
    """Create a mock DetectIntentResponse object for testing."""
    return types.session.DetectIntentResponse(query_result=mock_query_result)

@pytest.fixture
def mock_detect_intent_response_no_text_input(
    mock_query_result_params_no_text_input):
    """Create a mock DetectIntentResponse object for testing."""
    return types.session.DetectIntentResponse(
        query_result=mock_query_result_params_no_text_input)

@pytest.fixture
def mock_detect_intent_response_tools_playbooks_flows(
    mock_query_result_tools_playbooks_flows):
    """Create a mock DetectIntentResponse object for testing."""
    return types.session.DetectIntentResponse(
        query_result=mock_query_result_tools_playbooks_flows)

@pytest.fixture
def mock_environment_obj(test_config):
    """Create a mock Environment object for testing."""
    return types.Environment(
        name=test_config["environment_id"],
        display_name=test_config["environment_name"]
    )

@pytest.fixture(autouse=True)
def mock_sessions_client(test_config):
    """Fixture to create a mocked SessionsClient."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.sessions.services.sessions.SessionsClient") as mock_client, \
        patch("dfcx_scrapi.core.environments.Environments.__init__") as mock_env, \
        patch("dfcx_scrapi.core.tools.Tools.__init__") as mock_tools, \
        patch("dfcx_scrapi.core.playbooks.Playbooks.__init__") as mock_playbooks, \
        patch("dfcx_scrapi.core.flows.Flows.__init__") as mock_flows:
        
        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()
        mock_env.return_value = None
        mock_tools.return_value = None
        mock_playbooks.return_value = None
        mock_flows.return_value = None

        yield mock_client

@pytest.fixture
def mock_environments_client(test_config):
    """Fixture to create a mocked EnvironmentsClient."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.environments.services.environments.EnvironmentsClient") as mock_client:
        
        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_client


@pytest.fixture
def mock_tools_client(test_config):
    """Fixture to create a mocked ToolsClient."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.tools.services.tools.ToolsClient") as mock_client:
        
        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_client

@pytest.fixture
def mock_playbooks_client(test_config):
    """Fixture to create a mocked PlaybooksClient."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.playbooks.services.playbooks.PlaybooksClient") as mock_client:
        
        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_client

@pytest.fixture
def mock_flows_client(test_config):
    """Fixture to create a mocked FlowsClient."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.flows.services.flows.FlowsClient") as mock_client:
        
        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_client

@pytest.fixture
def mock_tools_instance(test_config):
    tools_instance = MagicMock()
    tools_instance.get_tools_map.return_value = {
        test_config["tool_id"]: "mock tool"
    }

    return tools_instance

def test_session_id_property_valid_id(test_config):
    """Test session_id property with a valid session ID."""
    session = Sessions(
        agent_id = test_config["agent_id"],
        session_id = test_config["session_id_plain"])
    assert session.session_id == test_config["session_id_plain"]

def test_printmd(test_config):
    session = Sessions()
    with patch("dfcx_scrapi.core.sessions.display") as mock_display:
        session.printmd("test string")
        mock_display.assert_called()

def test_build_query_input(test_config):
    session = Sessions()
    query_input = session._build_query_input(
        text=test_config["text"], language_code=test_config["language_code"]
    )
    assert isinstance(query_input, types.session.QueryInput)
    assert query_input.text.text == test_config["text"]
    assert query_input.language_code == test_config["language_code"]


def test_build_intent_query_input(test_config):
    session = Sessions()
    query_input = session.build_intent_query_input(
        intent_id=test_config["intent_id"], language_code=test_config["language_code"]
    )
    assert isinstance(query_input, types.session.QueryInput)
    assert query_input.intent.intent == test_config["intent_id"]
    assert query_input.language_code == test_config["language_code"]

def test_get_tool_action(test_config):
    session = Sessions()
    mock_tool_use = types.trace.ToolUse(action=test_config["tool_action"])
    action = session.get_tool_action(mock_tool_use)

    assert action == test_config["tool_action"]

def test_get_tool_params(test_config):
    session = Sessions()
    tool_use = types.ToolUse()
    tool_use.input_action_parameters = test_config["tool_input_params"]
    res = session.get_tool_params(tool_use.input_action_parameters)
    assert res == test_config["tool_input_params"]

def test_get_tool_params_empty():
    session = Sessions()
    params = {}
    res = session.get_tool_params(params)
    assert res == {}

def test_get_tool_params_nested(test_config):
    session = Sessions()
    tool_use = types.ToolUse()
    tool_use.input_action_parameters = test_config["tool_input_params_nested"]
    res = session.get_tool_params(tool_use.input_action_parameters)
    assert res == test_config["tool_input_params_nested"]

def test_get_playbook_name(test_config, monkeypatch):
    mock_playbooks_instance = MagicMock()
    mock_playbooks_instance.get_playbooks_map.return_value = {
        test_config["playbook_id"]: test_config["playbook_name"]
    }

    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "_playbooks_client", mock_playbooks_instance)
    
    name = session.get_playbook_name(playbook_id=test_config["playbook_id"])
    assert name == test_config["playbook_name"]


def test_get_tool_name(test_config, mock_tools_instance, monkeypatch):   
    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "_tools_client", mock_tools_instance)

    mock_tool_use = types.trace.ToolUse(tool=test_config["tool_id"])
    name = session.get_tool_name(tool_use=mock_tool_use)
    assert name == "mock tool"

def test_get_flow_name(test_config, monkeypatch):
    mock_flows_instance = MagicMock()
    mock_flows_instance.get_flows_map.return_value = {
        test_config["flow_id"]: "mock flow"
    }

    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "_flows_client", mock_flows_instance)

    name = session.get_flow_name(flow_id=test_config["flow_id"])
    assert name == "mock flow"

def test_collect_tool_responses(
        test_config, mock_tools_instance,
        mock_query_result_tools_playbooks_flows, monkeypatch):

    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "_tools_client", mock_tools_instance)
    tool_responses = session.collect_tool_responses(
        mock_query_result_tools_playbooks_flows)
    
    assert len(tool_responses) == 1
    assert tool_responses[0]["tool_name"] == "mock tool"
    assert tool_responses[0]["tool_action"] == test_config["tool_action"]
    assert tool_responses[0]["input_params"] == test_config["tool_input_params"]
    assert tool_responses[0]["output_params"] == test_config["tool_output_params"]

def test_collect_playbook_responses(
        test_config, mock_query_result_tools_playbooks_flows, monkeypatch):
    mock_playbooks_instance = MagicMock()
    mock_playbooks_instance.get_playbooks_map.return_value = {
        test_config["playbook_id"]: test_config["playbook_name"]
    }
    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "_playbooks_client", mock_playbooks_instance)
    playbook_responses = session.collect_playbook_responses(
        mock_query_result_tools_playbooks_flows)

    assert len(playbook_responses) == 4
    assert playbook_responses[0]["playbook_name"] == test_config["playbook_name"]

def test_collect_playbook_responses_no_playbook_invocation(
        test_config, monkeypatch):
    mock_playbooks_instance = MagicMock()
    mock_playbooks_instance.get_playbooks_map.return_value = {
        test_config["playbook_id"]: test_config["playbook_name"]
    }
    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "_playbooks_client", mock_playbooks_instance)

    mock_query_result = types.session.QueryResult(
    generative_info = types.session.GenerativeInfo(
        action_tracing_info=types.Example(
            actions=[],
           ),
        current_playbooks=[test_config["playbook_id"]]
        )
    )
    playbook_responses = session.collect_playbook_responses(mock_query_result)

    assert len(playbook_responses) == 0

def test_collect_flow_responses(test_config, monkeypatch):
    mock_flows_instance = MagicMock()
    mock_flows_instance.get_flows_map.return_value = {
        test_config["flow_id"]: "mock flow"
    }
    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "_flows_client", mock_flows_instance)
   
    mock_query_result = types.session.QueryResult(
        generative_info = types.session.GenerativeInfo(
            action_tracing_info=types.Example(
                actions=[
                    types.Action(
                        flow_invocation=types.trace.FlowInvocation(
                            flow=test_config["flow_id"]
                        ),
                    ),
                ],
            ),
        )
    )
    flow_responses = session.collect_flow_responses(mock_query_result)
    assert len(flow_responses) == 1
    assert flow_responses[0]["flow_name"] == "mock flow"

@patch("uuid.uuid4")
def test_build_session_id_no_environment(mock_uuid4, test_config):
    mock_uuid4.return_value = uuid.UUID("a1b2c3d4-e5f6-7890-1234-567890abcdef")

    session = Sessions()
    session_id = session.build_session_id(agent_id=test_config["agent_id"])

    assert session_id == test_config["session_id_plain"]
    assert session.session_id == test_config["session_id_plain"]

@patch("uuid.uuid4")
def test_build_session_id_with_environment(
    mock_uuid4, test_config, mock_environment_obj, monkeypatch):
    mock_uuid4.return_value = uuid.UUID("a1b2c3d4-e5f6-7890-1234-567890abcdef")

    mock_env_instance = MagicMock()
    mock_env_instance.get_environment_by_display_name.return_value = mock_environment_obj

    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "_env_client", mock_env_instance)

    session_id = session.build_session_id(
        agent_id=test_config["agent_id"],
        environment_name=test_config["environment_name"]
    )

    assert test_config["environment_id"] in session_id
    assert session.session_id == session_id
    assert session.session_id == test_config["session_id_with_env"]

def test_build_session_id_invalid_env(test_config, monkeypatch):
    mock_env_instance = MagicMock()
    mock_env_instance.get_environment_by_display_name.return_value = None

    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "_env_client", mock_env_instance)

    with pytest.raises(ValueError, match="Environment `BAD_ENV` does not exist."):
        session.build_session_id(
        agent_id=test_config["agent_id"],
        environment_name="BAD_ENV"
        )

@patch("uuid.uuid4")
def test_build_session_id_no_overwrite(mock_uuid4, test_config):
    mock_uuid4.return_value = uuid.UUID("a1b2c3d4-9999-9999-9999-567890abcdef")

    session = Sessions(session_id=test_config["session_id_plain"])
    session_id = session.build_session_id(agent_id=test_config["agent_id"], overwrite=False)

    assert session.session_id == test_config["session_id_plain"]
    assert session_id != session.session_id

def test_detect_intent(
    test_config, mock_sessions_client, mock_detect_intent_response
):
    mock_detect_intent = MagicMock(return_value=mock_detect_intent_response)
    mock_sessions_client.return_value.detect_intent = mock_detect_intent

    session = Sessions()
    query_result = session.detect_intent(
        agent_id=test_config["agent_id"],
        session_id=test_config["session_id_plain"],
        text=test_config["text"],
        language_code=test_config["language_code"],
    )

    assert query_result == mock_detect_intent_response.query_result
    assert isinstance(query_result, types.QueryResult)
    assert query_result.language_code == "en" #default lang code

    mock_detect_intent.assert_called_once()
    request = mock_detect_intent.call_args.kwargs['request']
    assert not request.query_params.end_user_metadata
    assert not request.query_params.parameters
    assert not request.query_params.time_zone

def test_detect_intent_params_and_end_user_metadata(
    test_config, mock_sessions_client, mock_detect_intent_response
):
    mock_detect_intent = MagicMock(return_value=mock_detect_intent_response)
    mock_sessions_client.return_value.detect_intent = mock_detect_intent

    session = Sessions()
    query_result = session.detect_intent(
        agent_id=test_config["agent_id"],
        session_id=test_config["session_id_plain"],
        text=test_config["text"],
        language_code=test_config["language_code"],
        parameters=test_config["parameters"],
        end_user_metadata=test_config["end_user_metadata"]
    )

    assert query_result == mock_detect_intent_response.query_result
    assert isinstance(query_result, types.QueryResult)
    assert query_result.language_code == "en" #default lang code
    
    mock_detect_intent.assert_called_once()
    request = mock_detect_intent.call_args.kwargs['request']
    assert request.query_params.end_user_metadata == test_config["end_user_metadata"]
    assert request.query_params.parameters == test_config["parameters"]
    assert not request.query_params.time_zone


def test_detect_intent_with_timezone(
    test_config, mock_sessions_client, mock_detect_intent_response
):
    mock_detect_intent = MagicMock(return_value=mock_detect_intent_response)
    mock_sessions_client.return_value.detect_intent = mock_detect_intent

    session = Sessions()
    query_result = session.detect_intent(
        agent_id=test_config["agent_id"],
        session_id=test_config["session_id_plain"],
        text=test_config["text"],
        language_code=test_config["language_code"],
        timezone="America/Los_Angeles"
    )

    assert query_result == mock_detect_intent_response.query_result
    assert isinstance(query_result, types.QueryResult)
    assert query_result.language_code == "en" #default lang code

    mock_detect_intent.assert_called_once()
    request = mock_detect_intent.call_args.kwargs['request']
    assert request.query_params.time_zone == "America/Los_Angeles"

# TODO (pmarlow): Tracking b/384222123 which causes Data Store Signals to
# "fail open", meaning they are always populated. Revise tests once this is
# resolved.
def test_detect_intent_with_data_store_signals(
    test_config, mock_sessions_client, mock_detect_intent_response
):
    mock_detect_intent = MagicMock(return_value=mock_detect_intent_response)
    mock_sessions_client.return_value.detect_intent = mock_detect_intent

    session = Sessions()
    query_result = session.detect_intent(
        agent_id=test_config["agent_id"],
        session_id=test_config["session_id_plain"],
        text=test_config["text"],
        language_code=test_config["language_code"],
        populate_data_store_connection_signals=True
    )

    assert query_result == mock_detect_intent_response.query_result

    mock_detect_intent.assert_called_once()
    request = mock_detect_intent.call_args.kwargs['request']
    assert request.query_params.populate_data_store_connection_signals

def test_detect_intent_with_intent_id(
    test_config, mock_sessions_client, mock_detect_intent_response
):
    mock_detect_intent = MagicMock(return_value=mock_detect_intent_response)
    mock_sessions_client.return_value.detect_intent = mock_detect_intent

    session = Sessions()
    query_result = session.detect_intent(
        agent_id=test_config["agent_id"],
        session_id=test_config["session_id_plain"],
        intent_id=test_config["intent_id"],
        language_code=test_config["language_code"]
    )

    assert query_result == mock_detect_intent_response.query_result

    mock_detect_intent.assert_called_once()
    request = mock_detect_intent.call_args.kwargs['request']
    assert request.query_input.intent.intent == test_config["intent_id"]

def test_detect_intent_invalid_session_id(test_config):
    session = Sessions()
    with pytest.raises(
        ValueError,
        match="Session ID must be provided in the following format:"):
        session.detect_intent(
           agent_id=test_config["agent_id"],
           session_id="invalid_session_id",
           text=test_config["text"],
           language_code=test_config["language_code"],
        )

def test_detect_intent_preset_parameters_no_text_input(
        test_config, mock_sessions_client,
        mock_detect_intent_response_no_text_input):
  mock_detect_intent = MagicMock(
      return_value=mock_detect_intent_response_no_text_input)
  mock_sessions_client.return_value.detect_intent = mock_detect_intent

  session = Sessions()
  query_result = session.detect_intent(
      agent_id=test_config["agent_id"],
      session_id=test_config["session_id_plain"],
      parameters=test_config["parameters"],
  )

  assert query_result == mock_detect_intent_response_no_text_input.query_result
  mock_detect_intent.assert_called_once()
  request = mock_detect_intent.call_args.kwargs['request']
  assert request.query_params.parameters == test_config["parameters"]
  assert not request.query_input.text.text

def test_preset_parameters_no_agent_id(test_config):
  session = Sessions()
  with pytest.raises(ValueError, match="resource_id must not be None"):
    session.detect_intent(
        agent_id=None,
        session_id=test_config["session_id_plain"],
        parameters=test_config["parameters"],
    )

def test_preset_parameters_no_session_id(test_config):
  session = Sessions()
  with pytest.raises(ValueError, match="Session ID must be provided in the following format:"):
    session.detect_intent(
        agent_id=test_config["agent_id"],
        session_id=None,
        parameters=test_config["parameters"],
    )

def test_get_agent_answer(test_config, mock_sessions_client, monkeypatch, mock_query_result_datastore):
    mock_detect_intent_response = MagicMock()
    mock_detect_intent_response.query_result = mock_query_result_datastore

    mock_detect_intent = MagicMock(return_value=mock_detect_intent_response)
    mock_sessions_client.return_value.detect_intent = mock_detect_intent

    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "build_session_id",
        MagicMock(return_value=test_config["session_id_plain"]))

    answer = session.get_agent_answer(user_query="who is the ceo?")
    assert answer == "Sundar Pichai is the CEO of Google.\nhttps://www.google.com (https://www.google.com)"
    mock_detect_intent.assert_called_once()

def test_get_agent_answer_no_citation(test_config, mock_sessions_client, monkeypatch):
    mock_detect_intent_response = MagicMock()
    mock_detect_intent_response.query_result = types.session.QueryResult(
        response_messages=[
            types.ResponseMessage(
                text=types.ResponseMessage.Text(
                    text=["Test Response"]
                    )
                    )
            ]
        )

    mock_detect_intent = MagicMock(return_value=mock_detect_intent_response)
    mock_sessions_client.return_value.detect_intent = mock_detect_intent

    session = Sessions(agent_id=test_config["agent_id"])
    monkeypatch.setattr(session, "build_session_id",
                        MagicMock(return_value=test_config["session_id_plain"]))

    answer = session.get_agent_answer(user_query="test query")
    assert answer == "Test Response ()"
    mock_detect_intent.assert_called_once()

def test_parse_result(
    test_config, mock_query_result_tools_playbooks_flows, monkeypatch
):
    session = Sessions(agent_id=test_config["agent_id"])
    mock_printmd = MagicMock()
    monkeypatch.setattr(session, "printmd", mock_printmd)

    session.parse_result(mock_query_result_tools_playbooks_flows)

    mock_printmd.assert_called()
    calls = mock_printmd.mock_calls

    tool_call_font = "<font color='dark red'>TOOL CALL:</font></b>"
    tool_res_font = "<font color='yellow'>TOOL RESULT:</font></b>"
    query_font = "<font color='green'><b> USER QUERY:</font></b>"
    response_font = "<font color='green'><b>AGENT RESPONSE:</font></b>"

    # Check for the user query
    assert any(query_font in call.args[0] for call in calls)
    assert any(tool_call_font in call.args[0] for call in calls)
    assert any(tool_res_font in call.args[0] for call in calls)
    assert any(response_font in call.args[0] for call in calls)



"""Methods to Generator Agent Tasks for arbitrary Agent inputs."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import random
from pathlib import Path
from typing import Any, Dict, List

from google.cloud.dialogflowcx_v3beta1 import types
from google.oauth2 import service_account
from vertexai.generative_models import GenerationConfig, GenerativeModel

from dfcx_scrapi.core.agents import Agents
from dfcx_scrapi.core.flows import Flows
from dfcx_scrapi.core.intents import Intents
from dfcx_scrapi.core.playbooks import Playbooks
from dfcx_scrapi.core.scrapi_base import ScrapiBase
from dfcx_scrapi.core.tools import Tools
from dfcx_scrapi.tools.gcs_utils import GcsUtils


class AgentTaskGenerator(ScrapiBase):

    def __init__(
        self,
        agent_id: str,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds: service_account.Credentials = None,
        scope=False,
        gcs: GcsUtils = None,
        debug: bool = False
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.agent_id = agent_id
        self.debug = debug
        if self.debug:
            self.set_logging_level("debug")

        self.agents = Agents(agent_id=self.agent_id, creds=self.creds)
        self.intents = Intents(agent_id=self.agent_id, creds=self.creds)
        self.flows = Flows(agent_id=self.agent_id, creds=self.creds)
        self.playbooks = Playbooks(agent_id=self.agent_id, creds=self.creds)
        self.tools = Tools(agent_id=self.agent_id, creds=self.creds)

        self.model = GenerativeModel(
            "gemini-1.5-flash-001",
            system_instruction=Prompts.task_system
            )

        self.gcs = gcs
        self.task_file = "agent_tasks.json"

        if self.gcs:
            self.short_id = self.agent_id.split("/")[-1]
            self.filename = f"testwiz/{self.short_id}/data/{self.task_file}"
            self.full_path = self.gcs.get_fully_qualified_path(self.filename)

    @staticmethod
    def write_agent_tasks_to_local(tasks: Dict[str, Any]):
        """Write Agent Task list to Local file."""
        task_file = "agent_tasks.json"
        data_dir = os.path.join(Path(__file__).resolve().parents[2], "data")
        cached_path = f"{data_dir}/{task_file}"

        # Write data back to file
        with open(cached_path, "w", encoding="UTF-8") as outfile:
            data_str = json.dumps(tasks)
            outfile.write(data_str)
            outfile.close()

    @staticmethod
    def get_sample_utterances(intent: Any):
        """Get 5 or less sample utterances from the intent."""
        num_to_select = min(len(intent.training_phrases), 5)
        intents = random.sample(intent.training_phrases, num_to_select)
        utterances = []

        for tp in intents:
            utterances.append(tp.parts[0].text)

        return utterances

    def gather_tool_spec_details(self, tool: types.Tool):
        """Determine Tool type and extract actions."""

        # Datastore Type
        connections = tool.data_store_spec.data_store_connections
        if len(connections) > 0:
            return tool.display_name, connections

        # Extension Type
        elif tool.extension_spec.name != "":
            return None, None

        # Function Call Type
        elif tool.function_spec.input_schema:
            input_schema = self.recurse_proto_marshal_to_dict(
                tool.function_spec.input_schema
            )
            return None, input_schema

        # OpenAPI Spec Type
        elif tool.open_api_spec.text_schema != "":
            return None, tool.open_api_spec.text_schema

    def _call_generate_content(self, prompt: str):
        return self.model.generate_content(
            prompt,
            safety_settings=self.build_safety_settings(),
            generation_config = GenerationConfig(
                temperature=0.1,
                top_p=0.95,
                candidate_count=1,
                max_output_tokens=8192,
                response_mime_type="application/json",
                response_schema=Schemas.task_schema,
                )
        )

    def load_or_create_task_file_gcs(self):
        """Load or create Agent Task file from GCS."""
        data = {"tasks": None}

        data_from_file = self.gcs.load_file_if_exists(
            bucket_name=self.gcs.bucket_name,
            filename=self.full_path
        )

        if data_from_file:
            logging.info("Loading Task List from GCS cache...")
            data = json.loads(data_from_file)

        else:
            self.gcs.write_dict_to_gcs(
                bucket_name=self.gcs.bucket_name,
                data=data,
                filename=self.full_path
            )

        return data

    def load_or_create_task_file_local(self):
        """Load or create Agent task file from Local."""
        data_dir = os.path.join(Path(__file__).resolve().parents[2], "data")
        cached_path = f"{data_dir}/{self.task_file}"
        data = {"tasks": None}

        # Create local if file doesn't exist
        if self.task_file not in os.listdir(data_dir):
            logging.info("Creating Agent Task File...")
            with open(cached_path, "w", encoding="UTF-8") as newfile:
                data_str = json.dumps(data)
                newfile.write(data_str)
                newfile.close()

        else:
            with open(cached_path, "r", encoding="UTF-8") as infile:
                data = json.load(infile)

        return data

    def write_agent_tasks_to_gcs(self, tasks: Dict[str, Any]):
        """Write Agent Task list to GCS file."""

        # write back to file
        self.gcs.write_dict_to_gcs(
            bucket_name=self.gcs.bucket_name,
            data=tasks,
            filename=self.full_path
        )

    def load_agent_tasks(self, force_reload: bool = False):
        """Check for cached Agent Task map or load new."""
        if self.gcs:
            data = self.load_or_create_task_file_gcs()
        else:
            data = self.load_or_create_task_file_local()

        tasks = data.get("tasks", None)

        # If the task metadata exists, but user wants to force reload anyways
        if tasks and force_reload:
            logging.info("Forcing reload of Agent Tasks...")
            tasks = self.get_agent_tasks()
            self.write_agent_tasks_to_local(tasks)
            if self.gcs:
                self.write_agent_tasks_to_gcs(tasks)

        # Check if tasks exists and if not, create
        if not tasks:
            logging.info("Agent Task list does not exist. Creating...")
            tasks = self.get_agent_tasks()

            if self.gcs:
                self.write_agent_tasks_to_gcs(tasks)
            else:
                self.write_agent_tasks_to_local(tasks)

        if isinstance(tasks, list):
            tasks = {"tasks": tasks}

        return tasks

    def trim_agent_tools(
            self, playbooks: List[Any], tools: List[Any]
            ) -> List[Any]:
        """Trim down the list of agent tools to only those that are ref'd."""
        playbook_tools = set()
        for playbook in playbooks:
            for tool in playbook.referenced_tools:
                playbook_tools.add(tool)

        tool_info = {}
        i = 0
        for tool in tools:
            if tool.name in playbook_tools:
                actions, schema = self.gather_tool_spec_details(tool)
                tool_info[f"tool_{i}"] = {
                    "name": tool.display_name,
                    "description": tool.description,
                    "actions": actions,
                    "schemas": schema
                }
            i += 1

        return tool_info

    def get_intent_info(self):
        """Get structured intent info from Agent to use in prompt."""
        EXCLUDE_INTENTS = [
            "Default Welcome Intent",
            "Default Negative Intent"
        ]

        all_intents = self.intents.list_intents(self.agent_id)
        intent_info = []
        for intent in all_intents:
            if intent.display_name in EXCLUDE_INTENTS:
                continue

            intent_info.append(
                {
                    "display_name": intent.display_name,
                    "description": intent.description,
                    "sample_utterances": self.get_sample_utterances(intent)
                }
            )

        return intent_info

    def get_playbook_steps(self, instruction: types.Playbook.Instruction):
        """Extract the playbook plain text steps."""
        all_steps = []
        for text in instruction.steps:
            all_steps.append(text.text)

        return all_steps

    def trim_agent_data(self, agent: Any):
        """Select only the relevant info from the agent proto."""
        return {
            "name": agent.display_name,
            "supported_languages": agent.supported_language_codes,
            "time_zone": agent.time_zone,
        }

    def trim_playbook_data(self, playbooks: Any):
        """Select only the relevant info from the playbook protos."""
        playbook_info = {}
        i = 0
        for playbook in playbooks:
            playbook_info[f"agent_{i}"] = {
                "name": playbook.display_name,
                "goal": playbook.goal,
                "instructions": self.get_playbook_steps(playbook.instruction)
            }
            i += 1

        return playbook_info

    def get_agent_tasks(self) -> Dict[str, Any]:
        """Extract App/Agent details and determine Agent's task list."""
        task_msg = "*** AUTO-GENERATING AGENT TASK LIST ***"
        logging.info(task_msg)

        agent = self.agents.get_agent(self.agent_id)
        agent = self.trim_agent_data(agent)

        playbooks = self.playbooks.list_playbooks(self.agent_id)
        tools = self.tools.list_tools(self.agent_id)
        tools = self.trim_agent_tools(playbooks, tools)

        # Order matters so that Tool trim can happen before playbook trim
        playbooks = self.trim_playbook_data(playbooks)
        intent_info = self.get_intent_info()
        flow_page_info = self.flows.get_flow_page_map(self.agent_id)

        prompt = Prompts.task_main.replace("{AGENT}", str(agent))
        prompt = prompt.replace("{PLAYBOOKS}", str(playbooks))
        prompt = prompt.replace("{TOOLS}", str(tools))
        prompt = prompt.replace("{INTENTS}", str(intent_info))
        prompt = prompt.replace("{FLOWS_AND_PAGES}", str(flow_page_info))

        if self.debug:
            logging.debug(prompt)

        res = self._call_generate_content(prompt)
        tasks_dict = json.loads(res.text)

        return tasks_dict

    def get_agent_tasks_from_user_input(
            self, tasks: Any) -> Dict[str, Any]:
        """Given an arbitrary user input, clean and format with LLM."""
        task_msg = "*** FORMATTING AGENT TASK LIST ***"
        logging.info(task_msg)

        prompt = Prompts.user_task_main.replace("{USER_DETAILS}", str(tasks))

        if self.debug:
            logging.debug(prompt)

        res = self._call_generate_content(prompt)
        tasks_dict = json.loads(res.text)

        return tasks_dict


class Prompts:
    # Main prompt template for task generation
    task_main: str = """**APP DETAILS**
{AGENT}

**AGENT DETAILS**
{PLAYBOOKS}
{FLOWS_AND_PAGES}
{INTENTS}

**TOOL DETAILS**
{TOOLS}
"""

    # System prompt for task generation
    task_system: str = """You are a senior virtual agent evaluator.
Your job is to determine the core tasks that a virtual agent can handle based on its provided details, playbook, and available tools.
Focus on the main functionalities that would be useful to an end user rather than individual steps or specific details within the process. Do not include generic functionalities such as "Intent Detection" or "Sentiment Analysis" unless those features provide a tangible outcome for the end user.  Explain each functionality clearly and concisely.
Return a concise list of the primary capabilities.

The information will be provided in a format like this:

**AGENT DETAILS**
<AGENT INFORMATION>

TOOL DETAILS:
<TOOL INFORMATION>

The resulting information should be returned in JSON format which will follow this schema:
```json
{
  "type": "object",
  "properties": {
    "tasks": {
      "type": "array",
      "items": {
        "type": "object",
        "properties": {
          "name": {
            "type": "string"
          },
          "description": {
            "type": "string"
          }
        },
        "required": [
          "name",
          "description"
        ]
      }
    }
  },
  "required": [
    "tasks"
  ]
}
```
""" # noqa: E501

    user_task_main: str = """APP DETAILS PROVIDED BY USER:
{USER_DETAILS}
"""

class Schemas:
    task_schema = {
            "type": "object",
            "properties": {
                "tasks": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": "string"
                                },
                            "description": {
                                "type": "string"
                                }
                            },
                        "required": [
                            "name",
                            "description"
                            ]
                    }
                }
            },
            "required": [
                "tasks"
                ]
            }



"""Functions to update the webhook env."""
import logging
import re

from dfcx_scrapi.core.webhooks import Webhooks

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
web=Webhooks()

def update_webhook(agent_id,env):
    """
    Updates the environment portion in the apigee webhook end point
    """
    weblist=web.list_webhooks(agent_id=agent_id)
    logging.info("got the webhooklist")

    for webhook in weblist:
        currenturi=webhook.generic_web_service.uri
        pattern = re.compile(r"\bdev\b")
        updateduri=re.sub(pattern, env, currenturi)
        webhook.generic_web_service.uri=updateduri
        kwargs={"generic_web_service":webhook.generic_web_service}
        web.update_webhook(webhook_id=webhook.name,
                           webhook_obj=webhook,**kwargs)
    logging.info("replaced dev to %s and updated all the webhook urls",env)



"""Evaluation tooling for Generative features in Agent Builder and DFCX."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from ast import literal_eval
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Dict, List

import numpy as np
import pandas as pd
from google.cloud.dialogflowcx_v3beta1 import types
from google.oauth2 import service_account
from tqdm import tqdm

from dfcx_scrapi.core.agents import Agents
from dfcx_scrapi.core.conversation_history import ConversationHistory
from dfcx_scrapi.core.playbooks import Playbooks
from dfcx_scrapi.core.scrapi_base import ScrapiBase
from dfcx_scrapi.core.sessions import Sessions
from dfcx_scrapi.core.tools import Tools
from dfcx_scrapi.tools.agent_response import AgentResponse
from dfcx_scrapi.tools.dataframe_functions import DataframeFunctions
from dfcx_scrapi.tools.metrics import build_metrics

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

@dataclass
class Interaction:
    actions: List[types.Action] = field(default_factory=list)
    generative_info: types.GenerativeInfo = None
    playbook_invocation: str = None
    query: str = None
    query_result: types.QueryResult = None
    responses: List = field(default_factory=list)
    tool_calls: List = field(default_factory=list)

    def __post_init__(self):
        if hasattr(self.query_result, "generative_info"):
            self.generative_info = self.query_result.generative_info
            self.actions = self.generative_info.action_tracing_info.actions

class Evaluations(ScrapiBase):
    """Evaluation tooling for Generative features in Agent Builder and DFCX."""

    def __init__(
        self,
        agent_id: str,
        creds_path: str = None,
        creds_dict: Dict[str, str] = None,
        creds: service_account.Credentials = None,
        metrics: List[str] = ["response_similarity"],
        debug: bool = False,
        generation_model: str = "gemini-1.5-flash-001",
        embedding_model: str = "text-embedding-004",
        playbooks_map: Dict[str, Any] = None,
        tools_map: Dict[str, Any] = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
        )

        self.agent_id = agent_id
        self.session_id = None

        print("Initializing Vertex AI...")
        self.init_vertex(self.agent_id)
        self.sessions_client = Sessions(
            agent_id=self.agent_id, tools_map=tools_map, creds=self.creds)
        self.playbooks_client = Playbooks(
            agent_id=self.agent_id,
            playbooks_map=playbooks_map, creds=self.creds)
        self.tools_client = Tools(
            agent_id=self.agent_id, tools_map=tools_map, creds=self.creds)
        self.ar = AgentResponse()
        self.ch = ConversationHistory(agent_id=self.agent_id, creds=self.creds)
        self.project_id = self.get_project_id_from_agent_id(
            resource_id=agent_id
            )
        self.location_id = self.get_location_id_from_agent_id(
            resource_id=agent_id
            )
        self.tools_map = None
        self.playbooks_map = None
        self.action_counter = 1
        self.model_id = generation_model

        self.embedding_model = self.model_setup(embedding_model)

        self.user_input_metrics = metrics
        self.metrics = build_metrics(
            metrics=self.user_input_metrics,
            genai_client=ScrapiBase._get_genai_client(
                project_id=self.project_id,
                location_id=self.location_id
                ),
            model_id=self.model_id,
            embedding_model=self.embedding_model
            )
        self.unexpected_rows = []

        if debug:
            logging.basicConfig(level=logging.DEBUG, force=True)
        if not debug:
            logging.basicConfig(level=logging.ERROR, force=True)

    @staticmethod
    def clean_outputs(df: pd.DataFrame) -> pd.DataFrame:
        """Clean final output dataframe."""
        # drop cols used for response mapping
        df = df.drop(columns=[
            "utterance_pair",
            "tool_pair",
            "playbook_pair",
            "flow_pair"
            ])
        value_map = {}
        for col, dtype in zip(df.columns, df.dtypes):
            if dtype in ["string", "object"]:
                value_map[col] = ""
            elif dtype == "float64":
                value_map[col] = np.nan

        df.fillna(value=value_map, inplace=True)

        return df

    @staticmethod
    def process_playbook_invocations(
        responses: List[str],
        index: int,
        row: pd.Series,
        df: pd.DataFrame) -> pd.DataFrame:
        if row["playbook_pair"] in [None, "", "NaN", "nan"]:
            playbook_index_list = [index]
        else:
            playbook_index_list = literal_eval(row["playbook_pair"])

        for idx in playbook_index_list:
            playbook = responses.pop(0)
            df.loc[int(idx), "res_playbook_name"] = playbook["playbook_name"]

        return df

    @staticmethod
    def process_flow_invocations(
        responses: List[str],
        index: int,
        row: pd.Series,
        df: pd.DataFrame) -> pd.DataFrame:
        if row["flow_pair"] in [None, "", "NaN", "nan"]:
            flow_index_list = [index]
        else:
            flow_index_list = literal_eval(row["flow_pair"])

        for idx in flow_index_list:
            flow = responses.pop(0)
            df.loc[int(idx), "res_flow_name"] = flow["flow_name"]

        return df

    @staticmethod
    def process_tool_invocations(
        tool_responses: List[Dict],
        index: int,
        row: pd.Series,
        df: pd.DataFrame,
    ) -> pd.DataFrame:
        """Process tool invocations and map them
        to the correct rows in the dataframe."""
        # Get the list of indices where tool responses should be mapped
        if row["tool_pair"] in [None, "", "NaN", "nan"]:
            tool_index_list = [index]
        else:
            tool_index_list = literal_eval(row["tool_pair"])

        # Process each tool response and map it to the corresponding index
        for i, idx in enumerate(tool_index_list):
            if i < len(tool_responses):
                tool = tool_responses[i]
                df.loc[int(idx), "res_tool_name"] = (
                    tool.get("tool_name", "")
                )
                df.loc[int(idx), "res_tool_action"] = (
                    tool.get("tool_action", "")
                )
                df.loc[int(idx), "res_input_params"] = (
                    str(tool.get("input_params", {}))
                )
                df.loc[int(idx), "res_output_params"] = (
                    str(tool.get("output_params", {}))
                )
            else:
                df.loc[int(idx), [
                    "res_tool_name",
                    "res_tool_action",
                    "res_input_params",
                    "res_output_params"
                    ]] = [
                    "NO_TOOL_RESPONSE",
                    "NO_TOOL_RESPONSE",
                    "NO_TOOL_RESPONSE",
                    "NO_TOOL_RESPONSE"
                ]

        return df

    @staticmethod
    def append_row(
        df_rows: List[pd.Series],
        eval_id: str,
        action_id: str,
        action_type: str,
        action_input: Any,
        action_input_parameters: str = "",
        tool_action: str = "",
        notes: str = ""):
        df_rows.append({
            'eval_id': eval_id,
            'action_id': action_id,
            'action_type': action_type,
            'action_input': action_input,
            'action_input_parameters': action_input_parameters,
            'tool_action': tool_action,
            'notes': notes
        })

    def parse_tool_use_from_conversation_history(
            self,
            tool_use: types.ToolUse) -> Dict[str, Any]:
        tool_name = self.tools_map.get(tool_use.tool, None)
        input_params = self.recurse_proto_marshal_to_dict(
            tool_use.input_action_parameters)

        # The DFCX proto for input params will add an extra top level key that
        # needs to be removed before writing to our output sheet. This only
        # applies to input_action_parameters
        if len(input_params) == 1 and input_params.get("", None):
            input_params = input_params.get("")

        output_params = self.recurse_proto_marshal_to_dict(
            tool_use.output_action_parameters)

        return {
            "tool_name": tool_name,
            "action": tool_use.action,
            "input_parameters": input_params,
            "output_parameters": output_params
        }

    def append_user_query(
            self,
            df_rows: List[Dict[str, Any]],
            eval_id: str,
            interaction: Interaction) -> None:
            self.append_row(
                df_rows=df_rows,
                eval_id=eval_id,
                action_id=self.action_counter,
                action_type="User Utterance",
                action_input=interaction.query
                )

            self.action_counter += 1

    def append_playbook(
            self,
            df_rows: List[Dict[str, Any]],
            eval_id: str,
            interaction: Interaction) -> int:
        if interaction.playbook_invocation:
            self.append_row(
                df_rows=df_rows,
                eval_id=eval_id,
                action_id=self.action_counter,
                action_type="Playbook Invocation",
                action_input=interaction.playbook_invocation
                )
            self.action_counter += 1

    def append_tools(
            self,
            df_rows: List[Dict[str, Any]],
            eval_id: str,
            interaction: Interaction) -> int:

        count = 0
        for tool_call in interaction.tool_calls:
            tool_name = tool_call.get('tool_name', '')
            tool_action = tool_call.get('action', '')
            input_params = str(tool_call.get('input_parameters', {}))
            self.append_row(
                df_rows=df_rows,
                eval_id=eval_id,
                action_id=self.action_counter + count,
                action_type="Tool Invocation",
                action_input=tool_name,
                action_input_parameters=input_params,
                tool_action=tool_action
            )
            count += 1
        self.action_counter += count

    def append_responses(
            self,
            df_rows: List[Dict[str, Any]],
            eval_id: str,
            interaction: Interaction) -> int:

        count = 0
        for response in interaction.responses:
            self.append_row(
                df_rows=df_rows,
                eval_id=eval_id,
                action_id=self.action_counter + count,
                action_type="Agent Response",
                action_input=str(response)
                )
            count += 1
        self.action_counter += count


    def parse_interactions_from_conversation_history(
                self,
                conversation: types.Conversation) -> List[Dict[str, Any]]:
            results = []

            # Load maps if they don't already exist
            if not self.tools_map:
                self.tools_map = self.tools_client.get_tools_map(self.agent_id)
            if not self.playbooks_map:
                self. playbooks_map = self.playbooks_client.get_playbooks_map(
                    self.agent_id)

            for conv_interaction in conversation.interactions:
                interaction = Interaction(
                    query=conv_interaction.request.query_input.text.text,
                    query_result=conv_interaction.response.query_result
                )

                if interaction.generative_info:
                    for action in interaction.actions:
                        if "tool_use" in action:
                            tool_calls = (
                                self.parse_tool_use_from_conversation_history(
                                    action.tool_use)
                            )
                            interaction.tool_calls.append(tool_calls)

                        elif "agent_utterance" in action:
                            response_text = action.agent_utterance.text
                            interaction.responses.append(response_text)

                        elif "playbook_invocation" in action:
                            playbook_name = self.playbooks_map.get(
                                action.playbook_invocation.playbook, None)
                            interaction.playbook_invocation = playbook_name

                results.append(interaction)
            results.reverse()

            return results

    def add_response_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()

        df.loc[:, "agent_response"] = pd.Series(dtype="str")
        df.loc[:, "agent_id"] = pd.Series(dtype="str")
        df.loc[:, "session_id"] = pd.Series(dtype="str")
        df.loc[:, "res_playbook_name"] = pd.Series(dtype="str")

        if "tool_call_quality" in self.user_input_metrics:
            df.loc[:, "res_tool_name"] = pd.Series(dtype="str")
            df.loc[:, "res_tool_action"] = pd.Series(dtype="str")
            df.loc[:, "res_input_params"] = pd.Series(dtype="str")
            df.loc[:, "res_output_params"] = pd.Series(dtype="str")

        return df

    def run_detect_intent_queries(
        self, df: pd.DataFrame, language_code: str = "en"
    ) -> pd.DataFrame:
        for index, row in tqdm(df.iterrows(), total=df.shape[0]):
            data = {}
            if row["action_id"] == 1:
                self.session_id = self.sessions_client.build_session_id(
                    self.agent_id)
                data["session_id"] = self.session_id
                data["agent_id"] = self.agent_id

            else:
                data["session_id"] = self.session_id
                data["agent_id"] = self.agent_id

            # If the incoming dataset has an empty value in the row, skip it
            # this is because we build the incoming dataset with multi-row
            # actions to be able to evaluate `inner-loop` tasks
            if row["action_type"] != "User Utterance":
                continue

            session_parameters = None

            if "session_parameters" in row:
                session_parameters = self.str_to_dict(row["session_parameters"])

            res = self.sessions_client.detect_intent(
                agent_id=self.agent_id,
                session_id=self.session_id,
                text=row["action_input"],
                parameters=session_parameters,
                language_code=language_code
            )
            # Add data to the existing row
            df.loc[index, ["session_id", "agent_id"]] = [
                data["session_id"],
                data["agent_id"],
            ]
            text_res = self.ar._extract_text(res)

            # Handle Agent Responses
            if row["utterance_pair"] != "":
                utterance_idx = int(row["utterance_pair"])
                df.loc[utterance_idx, ["agent_response"]] = [text_res]

            else:
                # collect the data for inserting later
                self.unexpected_rows.append(
                    {
                        "session_id": data["session_id"],
                        "agent_id": data["agent_id"],
                        "action_type": "UNEXPECTED Agent Response",
                        "index": index,
                        "column": "agent_response",
                        "data": text_res
                    }
                    )

            # Handle Playbook Invocations
            playbook_responses = (
                self.sessions_client.collect_playbook_responses(res)
            )
            if len(playbook_responses) > 0:
                df = self.process_playbook_invocations(
                    playbook_responses, index, row, df
                )

            # Handle Flow Invocations
            flow_responses = self.sessions_client.collect_flow_responses(res)
            if len(flow_responses) > 0:
                df = self.process_flow_invocations(
                    flow_responses, index, row, df
                )

            # Handle Tool Invocations
            if "tool_call_quality" in self.user_input_metrics:
                tool_responses = (
                    self.sessions_client.collect_tool_responses(res)
                )
                if tool_responses:  # Only call if not empty
                    df = self.process_tool_invocations(
                        tool_responses,
                        index,
                        row,
                        df
                    )

        return df

    def insert_unexpected_rows(self, df: pd.DataFrame) -> pd.DataFrame:
        """Insert any unexpected rows collected during runtime."""
        if self.unexpected_rows:
            for row in reversed(self.unexpected_rows):
                index = row["index"]
                new_row = pd.DataFrame(columns=df.columns, index=[index])
                new_row["session_id"] = row["session_id"]
                new_row["agent_id"] = row["agent_id"]
                new_row["action_type"] = row["action_type"]
                new_row[row["column"]] = row["data"]
                df = pd.concat(
                    [
                        df.iloc[:index],
                        new_row,
                        df.iloc[index:]
                    ])

        df = df.sort_index()

        return df

    def run_evals(self, df: pd.DataFrame) -> pd.DataFrame:
        print("Starting Evals...")

        for metric in self.metrics:
           df = pd.concat([df, metric.run(df)], axis=1)

        return df

    def scrape_results(
        self, df: pd.DataFrame, language_code: str = "en"
    ) -> pd.DataFrame:
        df = self.add_response_columns(df)
        df = self.run_detect_intent_queries(df, language_code=language_code)
        df = self.insert_unexpected_rows(df)

        return df

    def run_query_and_eval(
        self, df: pd.DataFrame, language_code: str = "en"
    ) -> pd.DataFrame:
        df = self.scrape_results(df, language_code=language_code)
        df = self.run_evals(df)
        df = self.clean_outputs(df)

        return df

    def create_dataset_from_conv_ids(
            self,
            conversation_ids: List) -> pd.DataFrame:
        columns = [
            'eval_id', 'action_id', 'action_type', 'action_input',
            'action_input_parameters', 'tool_action', 'notes'
        ]
        df_rows = []

        # Load maps if they don't already exist
        if not self.tools_map:
            self.tools_map = self.tools_client.get_tools_map(self.agent_id)
        if not self.playbooks_map:
            self. playbooks_map = self.playbooks_client.get_playbooks_map(
                self.agent_id)

        for idx, conv_id in enumerate(conversation_ids, start=1):
            eval_id = f"{idx:03d}"
            convo = self.ch.get_conversation(conv_id)
            interactions = self.parse_interactions_from_conversation_history(
                convo)

            self.action_counter = 1

            for interaction in interactions:
                self.append_user_query(df_rows, eval_id, interaction)
                self.append_playbook(df_rows, eval_id, interaction)
                self.append_tools(df_rows, eval_id, interaction)
                self.append_responses(df_rows, eval_id, interaction)

        return pd.DataFrame(df_rows, columns=columns)


class DataLoader:
    def __init__(
            self,
            creds_path: str = None,
            creds_dict: Dict[str, str] = None,
            creds: service_account.Credentials = None,
            agent_id: str = None,
            sheet_name: str = None,
            language_code: str = "en"
            ):

        self.agent_id = agent_id
        self.sheet_name = sheet_name
        self.dffx = DataframeFunctions(
            creds_path=creds_path, creds_dict=creds_dict, creds=creds
            )
        self.required_columns = [
            "eval_id",
            "action_id",
            "action_type",
            "action_input",
        ]
        self.language_code = language_code

    @staticmethod
    def get_matching_list_idx(a, b):
        """Helper method to find index pairs in the dataset.

        Compare lists and find the idx from list a where each element in b fits.
        This is used to determine exactly where the utterance or tool pairs
        exist in a given dataframe. The pairs are then used to determine where
        to write the results after the online inference is complete and the
        evals have been computed.
        """
        if not b:
            return [(a[0], [])]  # if b is empty, return

        result = []
        i, j = 0, 0

        current_b = []
        while i < len(a) and j < len(b):
            if a[i] < b[j]:
                current_a = a[i]
                if len(current_b) > 0:
                    result.append((a[i - 1], current_b))
                    current_b = []
                i += 1
            elif a[i] > b[j]:
                current_b.append(b[j])
                j += 1

        # if we're at end of list a, and still have list b
        # extend the remainder of b
        if i == len(a):
            current_b.extend(b[j:])
            result.append((current_a, current_b))

        # if we're at the end of list b, then append our current positions
        if j == len(b):
            result.append((current_a, current_b))

        return result

    @staticmethod
    def pair_utterances(df: pd.DataFrame) -> pd.DataFrame:
        "Identifies pairings of user_utterance and agent_utterance by eval_id."
        df["utterance_pair"] = pd.Series(dtype="string")
        grouped = df.groupby("eval_id")

        for _, group in grouped:
            user = group[
                group["action_type"] == "User Utterance"
            ].index.tolist()
            agent = group[
                group["action_type"] == "Agent Response"
            ].index.tolist()
            pairs = list(
                zip(user, agent)
            )

            # Create pairs of user/agent row indices
            for pair in pairs:
                df.loc[pair[0], "utterance_pair"] = str(pair[1])

        return df

    @staticmethod
    def get_agent_id_from_results(df: pd.DataFrame) -> pd.DataFrame:
        """Extract unique Agent ID from eval results."""
        agent_id_vals = df.agent_id.dropna().unique().tolist()
        for id in agent_id_vals:
            if len(id) > 0:
                return id

        return ""

    @staticmethod
    def get_model_name(settings: types.GenerativeSettings) -> str:
        """Get the model name from the Generative Settings."""
        model_name = settings.llm_model_settings.model
        model_map = {
            "gemini-pro": "gemini-1.0.pro-001",
            "gemini-1.5-pro": "gemini-1.5-pro-001",
            "gemini-ultra": "gemini-ultra",
            "text-unicorn-001": "text-unicorn-001",
            "gemini-1.5-flash": "gemini-1.5-flash-001",
            "text-bison-002": "text-bison-002"
        }

        return model_map.get(model_name, "")

    def pair_tool_calls(self, df: pd.DataFrame) -> pd.DataFrame:
        """Pairs user utterances with indices of relevant tool invocations."""

        df["tool_pair"] = pd.Series(dtype="string")
        grouped = df.groupby("eval_id")

        for _, group in grouped:
            tool_indices = []
            last_user_utterance_index = None

            for index, row in group.iterrows():
                if row["action_type"] == "User Utterance":
                    # Assign accumulated tool indices to
                    # the *previous* user utterance (if any)
                    if last_user_utterance_index is not None:
                        df.loc[last_user_utterance_index, "tool_pair"] = (
                            str(tool_indices)
                        )
                    # Reset for the current user utterance:
                    tool_indices = []
                    last_user_utterance_index = index

                elif row["action_type"] == "Tool Invocation":
                    tool_indices.append(index)

            # After processing the group, assign any remaining
            # tool indices to the last user utterance
            if last_user_utterance_index is not None and tool_indices:
                df.loc[last_user_utterance_index, "tool_pair"] = (
                    str(tool_indices)
                )

        return df

    def pair_playbook_calls(self, df: pd.DataFrame) -> pd.DataFrame:
        "Identifies pairings of agent_utterance/playbook_invocation by eval_id."
        df["playbook_pair"] = pd.Series(dtype="string")
        grouped = df.groupby("eval_id")

        for _, group in grouped:
            user = group[
                group["action_type"] == "User Utterance"
            ].index.tolist()
            playbook_list = group[
                group["action_type"] == "Playbook Invocation"
            ].index.tolist()

            pairs = self.get_matching_list_idx(
                user, playbook_list
            )

            # Create pairs of user/playbook_list row indices
            for pair in pairs:
                df.loc[pair[0], "playbook_pair"] = str(pair[1])

        return df

    def pair_flow_calls(self, df: pd.DataFrame) -> pd.DataFrame:
        "Identifies pairings of agent_utterance/flow_invocation by eval_id."
        df["flow_pair"] = pd.Series(dtype="string")
        grouped = df.groupby("eval_id")

        for _, group in grouped:
            user = group[
                group["action_type"] == "User Utterance"
            ].index.tolist()
            flow_list = group[
                group["action_type"] == "Flow Invocation"
            ].index.tolist()

            pairs = self.get_matching_list_idx(
                user, flow_list
            )

            # Create pairs of user/flow_list row indices
            for pair in pairs:
                df.loc[pair[0], "flow_pair"] = str(pair[1])

        return df

    def validate_input_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        """Validate input columns"""
        input_cols = set(df.columns.to_list())
        req_cols = set(self.required_columns)

        if not req_cols.issubset(input_cols):
            missing_cols = req_cols - input_cols
            raise ValueError(
                f"Missing columns: {missing_cols}. Required Columns are: "
                f"{self.required_columns}"
            )

        return df

    def check_existing_tab_name(
        self, sheet_name: str, results_tab: str
    ) -> bool:
        """Check to see if tab already exists."""
        sheet = self.dffx.sheets_client.open(sheet_name)
        existing_sheet = False
        worksheets = sheet.worksheets()
        for worksheet in worksheets:
            if worksheet.title == results_tab:
                existing_sheet = True

        return existing_sheet

    def create_sheet_tab(self, df: pd.DataFrame, results_tab: str):
        sheet = self.dffx.sheets_client.open(self.sheet_name)
        sheet.add_worksheet(results_tab, rows=df.shape[0], cols=df.shape[1])

    def write_eval_results_to_sheets(
        self, df: pd.DataFrame, sheet_name: str, results_tab: str = None
    ):
        tab_name_exists = self.check_existing_tab_name(sheet_name, results_tab)
        if results_tab and not tab_name_exists:
            self.create_sheet_tab(df, results_tab)
            self.dffx.dataframe_to_sheets(sheet_name, results_tab, df)

        elif results_tab and tab_name_exists:
            self.dffx.dataframe_to_sheets(sheet_name, results_tab, df)

        # auto generate a tab name and create it for the user
        else:
            today = datetime.today().strftime("%Y-%m-%d")
            results_tab = f"{today}-Eval Run"
            if not self.check_existing_tab_name(sheet_name, results_tab):
                self.create_sheet_tab(df, results_tab)

            self.dffx.dataframe_to_sheets(sheet_name, results_tab, df)


    def build_report_summary(self, df: pd.DataFrame) -> pd.DataFrame:
        # Check for agent_id or get from dataframe
        if not self.agent_id:
            self.agent_id = self.get_agent_id_from_results(df)

        # Get Generative Settings for report data
        a = Agents(language_code=self.language_code)
        agent = a.get_agent(self.agent_id)
        gen_settings = a.get_generative_settings(
            self.agent_id, language_code=self.language_code)
        model_name = self.get_model_name(gen_settings)

        current_datetime = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        metrics_info = {}
        if "similarity" in df.columns:
            metrics_info["similarity"] = df["similarity"].mean()
        if "tool_name_match" in df.columns:
            metrics_info["tool_match"] = df["tool_name_match"].mean()

        eval_results_summary = pd.DataFrame({
            'timestamp': [current_datetime],
            'total_conversations': [len(df['eval_id'].unique())],
            'model_name': [model_name],
            'agent_name': agent.display_name,
            'agent_id': [self.agent_id],
            'notes': [""]
        })

        # insert metrics for report
        insert_index = eval_results_summary.columns.get_loc(
            "total_conversations") + 1
        for metric, value in metrics_info.items():
            if (isinstance(value, float) and np.isnan(value)):
                value = "-"
            eval_results_summary.insert(insert_index, metric, [value])
            insert_index += 1

        return eval_results_summary

    def append_test_results_to_sheets(
        self, results: pd.DataFrame, sheet_name: str, summary_tab: str
        ):

        summary = self.build_report_summary(results)

        client = self.dffx.sheets_client
        gsheet = client.open(sheet_name)
        sheet = gsheet.worksheet(summary_tab)

        sheet.append_rows(
            summary.values.tolist(), value_input_option="USER_ENTERED"
        )

    def convert_column_types(self, df: pd.DataFrame) -> pd.DataFrame:
        """Convert column types as needed."""
        STR_COLUMNS = [
            "eval_id", "action_type", "action_input", "action_input_parameters",
            "tool_action", "notes", "session_parameters"
        ]

        for col in df.columns:
            if col in STR_COLUMNS and df[col].dtype != "object":
                df[col] = df[col].astype("object")

        return df


    def validate_and_prep_inputs(self, df: pd.DataFrame) -> pd.DataFrame:
        """Perform validations and transforms on input dataframe for evals."""
        # Check for action_id column, if none exists
        # add and assume all single turn queries
        if "action_id" not in df.columns.to_list():
            df["action_id"] = 1

        df["action_id"] = df["action_id"].astype(int)
        self.validate_input_columns(df)
        self.convert_column_types(df)

        df = self.pair_utterances(df)
        df = self.pair_tool_calls(df)
        df = self.pair_playbook_calls(df)
        df = self.pair_flow_calls(df)

        # fill remaining NA with empty string
        for col in df.columns:
            if df[col].dtype in ["object", "string"]:
                df[col] = df[col].fillna("")

        return df

    def from_google_sheets(
            self, sheet_name: str, sheet_tab: str) -> pd.DataFrame:
        """Load eval dataset from Google Sheets."""
        df = self.dffx.sheets_to_dataframe(sheet_name, sheet_tab)

        # Store sheet name for later use
        self.sheet_name = sheet_name
        df = self.validate_and_prep_inputs(df)

        return df

    def from_csv(self, file_path: str) -> pd.DataFrame:
        """Load eval dataset from local CSV file."""
        df = pd.read_csv(file_path)
        df = self.validate_and_prep_inputs(df)

        return df

    def from_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        """Load eval dataset from local premade dataframe."""
        df = self.validate_and_prep_inputs(df)

        return df



""" Deploy to prod functions """
import json
import logging
import sys

from shared.deployment import Deployment

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="PROD: %(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

def main(data):
    """
    Deploys and validates a Dialogflow CX agent in a production environment.
    This function orchestrates the deployment and validation of a Dialogflow CX
    agent in a production environment. It performs the following key steps:

    1. Imports the agent to the specified production webhook environment.
    2. Performs a language check on fulfillment entries, parameters, and routes,
        specifically for French Canadian ('fr-ca').
    3. Collects flow IDs.
    4. Manages version count and deletion.
    5. Cuts a new version of the agent.
    6. Deploys the new version to production.
    7. Synchronizes the agent between development and production environments.
    8. Updates the datastore with production information.

    Args:
        data: A dictionary containing configuration data, including the
            'prod_webhook_env' key.

    Raises:
        SystemExit: If the language check fails, indicating missing agent
            responses.
    """
    dep=Deployment(data)
    # call the steps sequentially
    dep.import_agent(webhookenv=data["prod_webhook_env"])

    entry,param,route,result=dep.fullfillment_lang_check(lang='fr-ca')

    logging.info("Entry fulfilment is %s",entry)
    logging.info("Param fulfilment is %s",param)
    logging.info("Route fulfilment is %s",route)
    if not result:
        print("some pages,parameters, routes does not have agent response")
        sys.exit(2)

    dep.collect_flow_id()
    dep.version_count_delete()
    dep.version_cut()
    dep.deploy_versions()
    dep.dev_prod_sync()
    dep.datastore_update("prod")



if __name__=='__main__':
    # read env variables
    with open("config.json" , encoding='utf-8') as config_file:
        config = json.load(config_file)

    SHA_ID=sys.argv[1]
    obj=f"UAT/{config['agent_name']}/{SHA_ID}"
    sha_agent_gcs_location=(
        f"gs://{config['bucket']}/UAT/{config['agent_name']}/{SHA_ID}"
    )
    logging.info("agent location %s", sha_agent_gcs_location)
    #adding additional variables to dict
    config["sha_agent_gcs_location"]=sha_agent_gcs_location
    config["target_project_id"] = config["prod_project"]
    config['target_environment_name']=config["prod_env_deploy"]
    with open("agent_artifacts/metadata.json",
              encoding='utf-8') as metadata_file:
        metadata = json.load(metadata_file)

    config["source_flow_names"]=metadata["source_flow_names"]
    config["updated_commit_message"]=metadata["updated_commit_message"]
    config["impacted_version_ids"]=metadata["impacted_version_ids"]

    # To execute steps in order
    main(config)



"""Utility and helper methods for using Google Cloud Storage."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from google.api_core.exceptions import NotFound
from google.cloud import storage


class GcsUtils:
    def __init__(self, gcs_path: str = None):
        self.client = storage.Client()

        if gcs_path:
            self.gcs_path = gcs_path
            self.bucket_name, self.file_path = self.get_bucket_name_and_path(
                gcs_path
                )
            self.bucket_exists(self.bucket_name)

    @staticmethod
    def get_bucket_name_and_path(gcs_path: str):
        """Strip gs:// or extract proper bucket name."""
        _, bucket_and_path = gcs_path.split("gs://")
        bucket_name, file_path = bucket_and_path.split("/", 1)

        return bucket_name, file_path

    def get_fully_qualified_path(self, filename: str):
        """Get the fully qualified path of the filename."""
        if self.file_path:
            return f"{self.file_path}/{filename}"
        else:
            return filename

    def read_file(self, filepath: str):
        """Read a file from GCS bucket."""

        # 1. Validate the file path
        if not filepath.startswith("gs://"):
            raise ValueError("Invalid filepath. Must start with 'gs://'")

        # 2. Split and extract components
        _, gcs_path = filepath.split("gs://")
        bucket_name, file_path = self.get_bucket_name_and_path(gcs_path)

        # 3. Access Google Cloud Storage
        bucket = self.client.bucket(bucket_name)
        blob = bucket.blob(file_path)

        # 4. Read the file content
        file_content = blob.download_as_text()

        return file_content

    def write_file(
            self,
            bucket_name: str,
            local_file_path: str,
            destination_name: str = None):
        """Write a file to GCS bucket."""

        # 1. Validate bucket name
        if not bucket_name:
            raise ValueError("Bucket name cannot be empty")

        # 2. Determine destination file path
        if destination_name:
            file_path = destination_name
        else:
            file_path = local_file_path.split("/")[-1]

        # 3. Access Google Cloud Storage
        bucket = self.client.bucket(bucket_name)
        blob = bucket.blob(file_path)

        # 4. Upload the file
        with open(local_file_path, "rb") as f:
            blob.upload_from_file(f)

        return f"File uploaded to gs://{bucket_name}/{file_path}"

    def write_dict_to_gcs(self, bucket_name: str, data: dict, filename: str):
        """Write a dict as a JSON file to a GCS bucket."""
        if not bucket_name or not filename:
            raise ValueError("Bucket name and filename cannot be empty")
        bucket = self.client.bucket(bucket_name)
        blob = bucket.blob(filename)
        json_string = json.dumps(data, indent=4)
        blob.upload_from_string(json_string, content_type="application/json")

    def load_file_if_exists(self, bucket_name: str, filename: str):
        bucket = self.client.bucket(bucket_name)
        blob = bucket.blob(filename)

        if blob.exists():
            full_path = f"gs://{bucket_name}/{filename}"
            return self.read_file(filepath=full_path)
        else:
            return None

    def bucket_exists(self, bucket_name: str):
        """Checks if a GCS bucket exists."""
        try:
            storage_client = storage.Client()
            storage_client.get_bucket(bucket_name)
        except NotFound:
            raise NotFound(f"GCS Bucket `{bucket_name}` does not exist.")



"""Test Cases Utility functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from typing import Dict, List, Union

from google.api_core import exceptions as core_exceptions
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf.struct_pb2 import Struct

from dfcx_scrapi.core import flows, intents, pages, scrapi_base, test_cases

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

class TestCasesUtil(scrapi_base.ScrapiBase):
    """Util class for CX test cases"""

    def __init__(
        self,
        agent_id: str,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.agent_id = agent_id
        self._global_test_cases = test_cases.TestCases(
            creds=self.creds, agent_id=self.agent_id)
        self.source_commons = {}
        self.target_commons = {}

    def _toggle_sentiment_in_convturns(
        self,
        conv_turns: List[types.ConversationTurn],
        sentiment: bool) -> List[types.ConversationTurn]:
        """Enable/disable enable_sentiment_analysis for all conversation turns.
        Args:
            conversation_turns: types.ConversationTurn
        Returns:
            a list of the conversation turns
        """
        new_conversation_turns = []
        for conversation_turn in conv_turns:
            conversation_turn.user_input.enable_sentiment_analysis = sentiment
            new_conversation_turns.append(conversation_turn)

        return new_conversation_turns

    def toggle_sentiment_test_cases(
        self,
        test_case_ids: List[str],
        sentiment: bool) -> List[types.TestCase]:
        """Enable/disable enable_sentiment_analysis for the given test case ids.
        Args:
            test_cases: a list of test case ids
            sentiment:
                if True then enable_sentiment_analysis = True
                else disable_sentiment_analysis = False
            agent_id: The formatted CX Agent ID
        Returns:
            a list of the updated test cases
        """
        test_cases_list = self._global_test_cases.list_test_cases(
            agent_id=self.agent_id, include_conversation_turns=True)
        updated_test_cases_list = []
        for test_case in test_cases_list:
            if test_case.name in test_case_ids:
                test_case.test_case_conversation_turns = (
                    self._toggle_sentiment_in_convturns(
                        conv_turns=test_case.test_case_conversation_turns,
                        sentiment=sentiment)
                )
                updated_test_cases_list.append(test_case)

        return updated_test_cases_list

    def _toggle_webhooks_in_convturns(
        self,
        conv_turns: List[types.ConversationTurn],
        is_webhook_enabled: bool) -> List[types.ConversationTurn]:
        """Sets the webhooks equal to True/False for all conversation turns.
        Args:
            conversation_turns: types.ConversationTurn
        Returns:
            conversation turns
        """

        new_conversation_turns = []
        for conversation_turn in conv_turns:
            conversation_turn.user_input.is_webhook_enabled = is_webhook_enabled
            new_conversation_turns.append(conversation_turn)

        return new_conversation_turns

    def toggle_webhooks_test_cases(
        self,
        test_case_ids: List[str],
        is_webhook_enabled: bool) -> List[types.TestCase]:
        """
        Enable/disable webhooks for the given test case ids.
        Args:
            test_cases: a list of the test case ids
            is_webhook_enabled: if enable then True else False
            agent_id: The formatted CX Agent ID
        Returns:
            list of the updated test caess
        """
        test_cases_list = self._global_test_cases.list_test_cases(
            agent_id=self.agent_id, include_conversation_turns=True)
        updated_test_cases_list = []
        for test_case in test_cases_list:
            if test_case.name in test_case_ids:
                test_case.test_case_conversation_turns = (
                    self._toggle_webhooks_in_convturns(
                        conv_turns=test_case.test_case_conversation_turns,
                        is_webhook_enabled=is_webhook_enabled)
                    )
                updated_test_cases_list.append(test_case)

        return updated_test_cases_list

    @staticmethod
    def _reverse_dict(dict_):
        """Utility method that reverse the dict keys and values
            Args:
                dict_: a dictionary
            Returns:
                a reversed dictionary
        """

        return {v: k for k, v in dict_.items()}

    def _get_resource_map(
        self,
        agent_id: str) -> Dict[str, Dict[str, str]]:
        """This method map flows, pages, and intents
            are in the dictionary format
          Arg:
            agent_id: The formatted CX Agent ID
          Returns:
            a dictionary of flows, pages, and intents
        """

        commons_config = {}
        commons_config["agent_id"] = agent_id
        dfcx_flows = flows.Flows(creds=self.creds, agent_id=agent_id)
        dfcx_pages = pages.Pages(creds=self.creds)
        dfcx_intents = intents.Intents(creds=self.creds, agent_id=agent_id)
        dfcx_testcases = test_cases.TestCases(
            creds=self.creds, agent_id=agent_id)
        dfcx_intents_map = dfcx_intents.get_intents_map(agent_id=agent_id)
        dfcx_flows_map = dfcx_flows.get_flows_map(agent_id=agent_id)
        dfcx_pages_map = {}
        for flow_id in dfcx_flows_map.keys():
            dfcx_pages_map[flow_id] = dfcx_pages.get_pages_map(flow_id=flow_id)
        commons_config["test_cases"] = dfcx_testcases
        commons_config["flows"] = dfcx_flows
        commons_config["pages"] = dfcx_pages
        commons_config["intents"] = dfcx_intents
        commons_config["flows_map"] = dfcx_flows_map
        commons_config["pages_map"] = dfcx_pages_map
        commons_config["intents_map"] = dfcx_intents_map
        commons_config["flows_map_reverse"] = (
            self._reverse_dict(dict_=commons_config["flows_map"]))
        commons_config["intents_map_reverse"] = (
            self._reverse_dict(dict_=commons_config["intents_map"]))
        commons_config["pages_map_reverse"] = (
            {f_id:self._reverse_dict(dict_=pages_map) \
             for f_id, pages_map in dfcx_pages_map.items()})

        return commons_config

    def _get_new_test_config(
        self,
        source_resource_map: dict,
        target_resource_map: dict,
        test_case: types.TestCase) -> Union[types.TestConfig, None]:
        """Update the test case types.test_config to the new types.test_config
            Update the uuids of flow, page, intent by the display_name
            If a TestConfig is found, it will return that,
            otherwise an empty TestConfig is returned.
          args:
            source_resource_map: a source resource dictionary
            target_resource_map: a target resource dictionary
            test_case: types.test_config object
          returns:
            types.test_config
        """
        source_flow_id = (
            # pylint: disable=W0212
            source_resource_map["test_cases"]._get_flow_id_from_test_config(
            test_case=test_case)
        )
        source_page_id = (
            # pylint: disable=W0212
            source_resource_map["test_cases"]._get_page_id_from_test_config(
                test_case=test_case, flow_id=source_flow_id)
        )
        source_flow = self._get_flow_display_name_by_id(
            source_flow_id, source_resource_map["flows_map"])
        source_page = self._get_page_display_name_by_id(
            source_flow_id, source_page_id, source_resource_map["pages_map"])
        target_test_config = types.TestConfig()

        if test_case.test_config.flow:
            target_flow_id = self._get_flow_id_by_flow_name(
                flow_name=source_flow,
                flows_map=target_resource_map["flows_map_reverse"])
            if target_flow_id:
                target_test_config.flow = target_flow_id
            else:
                return None

        elif test_case.test_config.page:
            target_flow_id = self._get_flow_id_by_flow_name(
                flow_name=source_flow,
                flows_map=target_resource_map["flows_map_reverse"])
            target_page_id = self._get_page_id_by_page_name(
                flow_id=target_flow_id,
                page_name=source_page,
                pages_map=target_resource_map["pages_map_reverse"])
            if target_page_id:
                target_test_config.page = target_page_id
            else:
                return None

        return target_test_config

    def _get_flow_display_name_by_id(
        self,
        flow_id: str,
        flows_map: Dict[str, str]) -> str:
        """Attempt to get the Flow Display Name from the Flows Map
            by the flow ID."""

        flow = flows_map.get(flow_id, None)

        return flow

    def _get_page_display_name_by_id(
        self,
        flow_id: str,
        page_id: str,
        pages_map: Dict[str, Dict[str, str]]) -> str:
        """Get the Page Display Name from the Pages Map based on the Page ID."""

        page_map = pages_map.get(flow_id, None)
        page = "START_PAGE"

        if page_map:
            page = page_map.get(page_id, None)

        return page

    def _get_flow_id_by_flow_name(
        self,
        flow_name: str,
        flows_map: Dict[str, str]) -> str:
        """Attempt to get the Flow id from the Flows Map
            by the flow display name."""

        return flows_map.get(flow_name, None)

    def _get_page_id_by_page_name(
        self,
        flow_id: str,
        page_name: str,
        pages_map: Dict[str, Dict[str, str]]) -> str:
        """Attempt to get the Page id from the Pages Map
            by the Page display name."""

        if not flow_id:
            return None
        page_map = pages_map.get(flow_id, None)
        page_id = None
        if  page_name in [
            "Start Page",
            "End Flow",
            "End Session",
            "End Flow With Failure",
            "End Flow With Human Escalation",
            "End Flow With Cancellation"
            ]:
            page_name = page_name.upper().replace(" ", "_")
            page_id = f"{flow_id}/pages/{page_name}"
        elif page_map:
            page_id = page_map.get(page_name, None)

        return page_id

    def _get_intent_id_by_intent_name(
        self,
        intent_name: str,
        intents_map: Dict[str, str]) -> str:
        """Attempt to get the intent id from the Intents Map
            by the intent's display name."""

        return intents_map.get(intent_name, None)

    def _get_intent_name_by_intent_id(
        self,
        intent_id: str,
        intents_map: Dict[str, str]) -> str:
        """Attempt to get the intent display name from the Intents Map
            by the intent's display name."""

        return intents_map.get(intent_id, None)

    def _get_intent_id_from_virtual_agent_output(
        self,
        virtual_agent_output: types.ConversationTurn.VirtualAgentOutput,
        target_resource_map: dict) -> Union[str, None]:
        """Attempt to find the intent id from the source's agent
        conversation_turn.virtual_agent_output triggered intent
        display name using the target's intents map
        """

        source_intent = (
          virtual_agent_output.triggered_intent.display_name
        )
        return self._get_intent_id_by_intent_name(
            intent_name=source_intent,
            intents_map=target_resource_map["intents_map_reverse"])

    def _get_page_id_from_virtual_agent_output(
        self,
        virtual_agent_output: types.ConversationTurn.VirtualAgentOutput,
        source_resource_map: dict,
        target_resource_map: dict) -> Union[str, None]:
        """Attempt to find the page id from the source's agent conversation turn
          virtual agent output page id to the target's pages map
        """

        source_page = virtual_agent_output.current_page.display_name
        source_page_id = virtual_agent_output.current_page.name
        source_flow = self._get_flow_display_name_by_id(
            flow_id = source_page_id.split("/pages/")[0],
            flows_map = source_resource_map["flows_map"])
        target_flow_id = self._get_flow_id_by_flow_name(
            flow_name=source_flow,
            flows_map=target_resource_map["flows_map_reverse"])
        target_page_id = self._get_page_id_by_page_name(
            flow_id=target_flow_id,
            page_name=source_page,
            pages_map=target_resource_map["pages_map_reverse"])

        return target_page_id

    def _get_intent_id_from_user_input(
        self,
        user_input: types.ConversationTurn.UserInput,
        source_resource_map: dict,
        target_resource_map: dict) -> Union[str, None]:
        """Attempt to find the intent id from the source's agent
          conversation turn user input intent id to the target's pages map
        """

        source_intent_id = user_input.input.intent
        source_intent = self._get_intent_name_by_intent_id(
            intent_id=source_intent_id,
            intents_map=source_resource_map["intents_map"])
        target_intent_id = self._get_intent_id_by_intent_name(
            intent_name=source_intent,
            intents_map=target_resource_map["intents_map_reverse"])

        return target_intent_id

    def _build_conversation_turns_for_target_agent(
        self,
        source_resource_map: dict,
        target_resource_map: dict,
        conversation_turns: types.ConversationTurn
        ) -> Union[List[types.ConversationTurn], None]:
        """ Update the uuids of the current page and intent
            by display name in each conversation turn.
          Args:
            source_resource_map: a source resource dictionary
            target_resource_map: a target commons_config dictionary
            conversation_turns: types.ConversationTurn
          Returns:
            types.test_case_conversation_turns
        """

        new_conversation_turns = []
        for conv_turn in conversation_turns:
            user_input_intent = conv_turn.user_input
            virtual_agent_output = conv_turn.virtual_agent_output
            if user_input_intent.input.intent:
                conv_turn.user_input.input.intent = (
                    self._get_intent_id_from_user_input(
                        user_input=user_input_intent,
                        source_resource_map=source_resource_map,
                        target_resource_map=target_resource_map
                        )
                )
                if not conv_turn.user_input.input.intent:
                    return None
            if virtual_agent_output.current_page:
                conv_turn.virtual_agent_output.current_page.name = (
                    self._get_page_id_from_virtual_agent_output(
                        virtual_agent_output=virtual_agent_output,
                        source_resource_map=source_resource_map,
                        target_resource_map=target_resource_map
                        )
                )
                if not conv_turn.virtual_agent_output.current_page.name:
                    return None
            if virtual_agent_output.triggered_intent:
                conv_turn.virtual_agent_output.triggered_intent.name = (
                    self._get_intent_id_from_virtual_agent_output(
                        virtual_agent_output=virtual_agent_output,
                        target_resource_map=target_resource_map
                    )
                )
                if not conv_turn.virtual_agent_output.triggered_intent.name:
                    return None
            new_conversation_turns.append(conv_turn)

        return new_conversation_turns

    def _log_missing_attributes(
        self,
        test_case: types.TestCase,
        test_config: types.TestConfig,
        conv_turns: List[types.ConversationTurn],
        last_test_conv_turns: List[types.ConversationTurn]):
        """This method is a logger that indicates which attributes are not
        convertible in the test case.
        Args:
          test_config: types.TestConfig
          conv_turns: List[types.ConversationTurn]
          last_test_conv_turns: List[types.ConversationTurn]
        """

        if test_config is None:
            logging.warning(
                "-- ERROR --  DFCX Test case failed to convert -- "
                f"test_case: {test_case.display_name} "
                "Reason: test_config is None.")
        if conv_turns is None:
            logging.warning(
                "-- ERROR --  DFCX Test case failed to convert -- "
                f"test_case: {test_case.display_name} "
                "Reason: test_conversation_turns is None.")
        if last_test_conv_turns is None:
            logging.warning(
                "-- ERROR --  DFCX Test case failed to convert -- "
                f"test_case: {test_case.display_name} "
                "Reason: last_test_result is None.")

    def convert_test_case(
        self,
        test_case: types.TestCase,
        source_agent_id: str,
        target_agent_id: str,
        rate_limit: int = 5) -> Union[types.TestCase, None]:
        """This method converts the test case from the source agent to the
          target agent.
          Note: In dfcx_scrapi.core.test_case.TestCases, it requires
          list_test_cases(include_conversation_turns=True),
          Otherwise, it will fail to convert.
        Args:
          test_case: types.TestCase
          source_agent_id: str
          target_agent_id: str
          rate_limit: int
        Returns:
          types.TestCase or None if  fails
        """

        if not test_case.test_case_conversation_turns:
            raise UserWarning(
                "types.TestCase is not convertible if "
                "test_case_conversation_turns is empty "
                f"test case : {test_case.display_name}"
            )
        if not self.source_commons.get("agent_id") == source_agent_id:
            self.source_commons = self._get_resource_map(
                agent_id=source_agent_id)
        if not self.target_commons.get("agent_id") == target_agent_id:
            self.target_commons = self._get_resource_map(
                agent_id=target_agent_id)

        test_config = self._get_new_test_config(
            self.source_commons, self.target_commons, test_case)
        conv_turns = self._build_conversation_turns_for_target_agent(
            self.source_commons,
            self.target_commons,
            test_case.test_case_conversation_turns)
        last_test_conv_turns = self._build_conversation_turns_for_target_agent(
            self.source_commons,
            self.target_commons,
            test_case.last_test_result.conversation_turns)

        if not all([test_config, conv_turns, last_test_conv_turns]):
            self._log_missing_attributes(
                test_case, test_config, conv_turns, last_test_conv_turns)
            return None

        test_case.name = None
        test_case.test_config = test_config
        test_case.test_case_conversation_turns = conv_turns
        test_case.last_test_result.conversation_turns = last_test_conv_turns
        new_test_case = None
        try:
            new_test_case = self.target_commons["test_cases"].create_test_case(
                test_case=test_case)
            time.sleep(rate_limit)
        except core_exceptions.InternalServerError as err:
            logging.error(
                f"-- ERROR -- InternalServerError caught on CX.detect -- {err}")
            logging.error("test_case: %s", test_case.display_name)
        except core_exceptions.ClientError as err:
            logging.error(
                f"-- ERROR -- ClientError caught on CX.detect -- {err}")
            logging.error("test_case: %s", test_case.display_name)
        logging.info(
            "-- SUCCESS -- DFCX test_case converted -- "
            f"test case: {new_test_case.display_name}")

        return new_test_case

    def _initialize_conversation_turn(self) -> types.ConversationTurn:
        """it initializes the conversation turn.
        Returns:
          types.ConversationTurn
        """
        conversation_turn = types.ConversationTurn()
        conversation_turn.user_input = (
            conversation_turn.UserInput(input=types.QueryInput()))

        return conversation_turn

    def _set_query_input(
        self,
        send_obj: Dict,
        conversation_turn: types.ConversationTurn
        ) -> types.ConversationTurn:
        """it sets the query input. In query input, it handle 4 types
          of input: text, event, intent, dtmf if not, then it raises a warning.
        Args:
          conversation_turn: types.ConversationTurn
        Returns:
          types.ConversationTurn
        """
        if "dtmf" in send_obj:
            finish_digit = send_obj.get("finish_digit", None)
            if finish_digit:
                finish_digit = str(finish_digit)
            conversation_turn.user_input.input.dtmf = (
                types.DtmfInput(
                    digits=send_obj["dtmf"],
                    finish_digit=finish_digit))
        elif "event" in send_obj:
            conversation_turn.user_input.input.event = (
                types.EventInput(event=send_obj["event"]))
        elif "text" in send_obj:
            conversation_turn.user_input.input.text = (
                types.TextInput(text=send_obj["text"]))
        elif "intent" in send_obj:
            conversation_turn.user_input.input.intent = (
                types.IntentInput(intent=send_obj["intent"]))
        else:
            raise UserWarning(
                "send_obj doesn't contain the proper query input type. "
                "it must contains one of [text, event, intent, dtmf] as a key")

        return conversation_turn

    def _convert_send_objs_to_conv_turns(
        self,
        send_objs: List[Dict],
        webhooks: bool) -> List[types.ConversationTurn]:
        """This function converts the send_objs to conversation turns.
          Iterate send_objs, set the params and a user_input of every turn.
        Args:
          send_objs: List[dict]
          webhooks: bool
        Returns:
          List[types.ConversationTurn]
        """
        conversation_turns = []
        for send_obj in send_objs:
            conversation_turn = self._initialize_conversation_turn()
            if "params" in send_obj:
                param = Struct()
                param.update(send_obj["params"])
                conversation_turn.user_input.injected_parameters = param
            conversation_turn.user_input.is_webhook_enabled = webhooks
            conversation_turn = self._set_query_input(
                send_obj=send_obj,
                conversation_turn=conversation_turn)
            conversation_turns.append(conversation_turn)
        return conversation_turns

    def _set_test_config(
        self,
        current_page) -> types.TestConfig:
        """it sets the test config of the test case.
        if current page is None, then it starts at Default Start Flow/Start page
        if current page is a page id, then it starts at the specific page id
        if current page is a flow id, then it starts at the flow start page
        Args:
          current_page: str
        Returns:
          types.TestConfig
        """
        test_config = types.TestConfig()
        if current_page:
            if "pages" in current_page:
                test_config.page = current_page
            else:
                test_config.flow = current_page

        return test_config

    def create_test_case_by_send_objs(
        self,
        display_name: str,
        send_objs: List[Dict],
        webhooks: bool = False,
        current_page: str = None,
        tags: List[str] = None,
        rate_limit: int = 5) ->  Union[types.TestCase, None]:
        """it creates a test case by a list of send_objs. Each send_obj
        consists of params and user input of every turn. Send_obj is commonly
        used in the e2e testing, dfcx_scrapi.core.conversation.reply().
        With this function, it can simultaneously create the test case while
        running the e2e test.
        Args:
          display_name: a display_name of the test case
          send_objs: a list of send_objs
          webhooks: False by default, if True, then it enables webhooks
          current_page: page_id of where the test case begins
          tags: List[str]
          rate_limit: int
        Returns:
          types.TestCase or None if fails
        """
        test_case = types.TestCase()
        test_case.display_name = display_name
        test_case.tags = tags
        if current_page:
            test_case.test_config = self._set_test_config(
                current_page=current_page)
        test_case.test_case_conversation_turns = (
            self._convert_send_objs_to_conv_turns(
                send_objs=send_objs, webhooks=webhooks)
        )
        response = None
        try:
            response = self._global_test_cases.create_test_case(
                test_case=test_case)
            time.sleep(rate_limit)
        except core_exceptions.InternalServerError as err:
            logging.error(
                f"-- ERROR -- InternalServerError caught on CX.detect -- {err}")
            logging.error("test_case: %s", test_case.display_name)
        except core_exceptions.ClientError as err:
            logging.error(
                f"-- ERROR -- ClientError caught on CX.detect -- {err}")
            logging.error("test_case: %s", test_case.display_name)

        return response



""" export functions"""

import json
import logging
import sys

from google.cloud import storage

from dfcx_scrapi.core.agents import Agents

from .flow_impacted import Impacted

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="dev: %(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

def agent_to_gcs(
    agent_name,
    project_id,
    environment_name,
    gsloc):
    """Exports a Dialogflow CX agent to Google Cloud Storage (GCS).

    This function exports a specified Dialogflow CX agent and its environment
    to a designated location in Google Cloud Storage.

    Args:
        agent_name: The display name of the agent to export.
        project_id: The ID of the Google Cloud project where the agent resides.
        environment_name: The display name of the environment to export.
        gsloc: The GCS bucket URI where the agent will be exported.

    Returns:
        None
    """
    agents=Agents()

    agent_details=agents.get_agent_by_display_name(
        display_name=agent_name,
        project_id=project_id
    )

    agent_id=agent_details.name
    agent_gcs_location=f"{gsloc}/{agent_name}"

    #export will replace exisitng agent in bucket
    agents.export_agent(agent_id=agent_id,
                        gcs_bucket_uri=agent_gcs_location,
                        environment_display_name=environment_name)
    logging.info("Agent export from dev done")

def meta_to_gcs(
    config_data,
    flow_names,
    commit_message,
    gsloc,
    agent_name,
    gcs_bucket,
    version_ids
    ):
    """Exports metadata to a JSON file in Google Cloud Storage (GCS).

    This function takes configuration data, flow names, a commit message,
    GCS location information, agent name, and version IDs, and creates a JSON
    file containing this metadata in the specified GCS bucket.

    Args:
        config_data: A dictionary containing configuration data.
        flow_names: A list of flow names.
        commit_message: The commit message to include in the metadata.
        gsloc: The full GCS URI where the metadata file will be stored.
        agent_name: The name of the agent.
        gcs_bucket: The name of the GCS bucket.
        version_ids: A list of version IDs.

    Returns:
        None
    """

    config_data["source_flow_names"]=flow_names
    config_data["impacted_version_ids"]= version_ids
    config_data["updated_commit_message"]=commit_message
    gcslist=gsloc.split("/")
    obj="/".join(gcslist[3:])

    bucket_obj = storage.Client().get_bucket(gcs_bucket)

    blob = bucket_obj.blob(f"{obj}/{agent_name}_metadata.json")
    blob.upload_from_string(data=json.dumps(config_data),
                            content_type='application/json')


if __name__=='__main__':
    # read env variables
    with open("config.json", encoding='utf-8') as config_file:
        config = json.load(config_file)

    source_project_id=config["dev_project"]
    source_agent_name=config["agent_name"]
    source_environment_name=config["dev_env_pull"]
    bucket=config["bucket"]
    user_commit_message=sys.argv[1]
    userid=sys.argv[2]
    updated_commit_message = (
    f"{user_commit_message} by {userid} "
    f"for {source_agent_name}"
    )
    impflows=Impacted(source_project_id=source_project_id,
                      source_agent_name=source_agent_name,
                      environment_name=source_environment_name)
    imp_flow_map,impacted_version_ids=impflows.check_flow()
    source_flow_names=list(imp_flow_map.values())
    source_flow_ids=list(imp_flow_map.keys())
    gs_loc=f"gs://{bucket}/exports/dev"

    logging.info("impacted flow is %(imp_flow_map)s"
                 , {'imp_flow_map': imp_flow_map})


    #Execute in steps
    agent_to_gcs(source_agent_name,
                 source_project_id,
                 source_environment_name,
                 gs_loc)
    meta_to_gcs(config,source_flow_names,
                updated_commit_message,gs_loc,
                source_agent_name,bucket,impacted_version_ids)



"""Test Class for Dataframe Function Methods in SCRAPI."""

# pylint: disable=redefined-outer-name

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock

import pytest
from google.oauth2.service_account import Credentials

from dfcx_scrapi.tools.dataframe_functions import DataframeFunctions


@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    email = "mock_email@testing.com"
    creds_path = "/Users/path/to/creds/credentials.json"
    creds_dict = {
        "type": "service_account",
        "project_id": project_id,
        "private_key_id": "1234",
        "private_key": "mock_key",
        "client_email": f"mock-account@{project_id}.iam.gserviceaccount.com",
        "client_id": "1234",
        "universe_domain": "googleapis.com",
    }

    mock_signer = MagicMock()
    mock_signer.key_id = "mock_key_id"
    mock_signer.sign.return_value = b"mock_signature"

    creds_object = Credentials(
        signer=mock_signer,
        token_uri="mock_token_uri",
        service_account_email=email,
        project_id=project_id,
        quota_project_id=project_id,
        scopes=[],
    )

    return {
        "project_id": project_id,
        "creds_path": creds_path,
        "creds_dict": creds_dict,
        "creds_object": creds_object,
    }


@pytest.fixture
def mock_dffx_setup(monkeypatch, test_config):
    """Fixture to create mock DataframeFunctions object w/ mocked clients."""

    # mocking all other classes used by DFFX
    mock_credentials_from_file = MagicMock(
        return_value=test_config["creds_object"]
    )

    monkeypatch.setattr(
        "google.oauth2.service_account.Credentials.from_service_account_file",
        mock_credentials_from_file,
    )

    # mocking all other classes used by DFFX
    def mock_scrapi_base_init(self, *args, **kwargs):
        # Simulate the original behavior
        if kwargs.get("creds_path"):
            self.creds = Credentials.from_service_account_file(
                kwargs.get("creds_path")
            )
        elif kwargs.get("creds_dict"):
            self.creds = Credentials.from_service_account_info(
                kwargs.get("creds_dict")
            )
        else:
            self.creds = kwargs.get("creds")

    def mock_entities_init(self, *args, **kwargs):
        pass

    def mock_intents_init(self, *args, **kwargs):
        pass

    def mock_flows_init(self, *args, **kwargs):
        pass

    def mock_pages_init(self, *args, **kwargs):
        pass

    def mock_route_groups_init(self, *args, **kwargs):
        pass

    monkeypatch.setattr(
        "dfcx_scrapi.core.scrapi_base.ScrapiBase.__init__",
        mock_scrapi_base_init,
    )

    monkeypatch.setattr(
        "dfcx_scrapi.core.entity_types.EntityTypes.__init__", mock_entities_init
    )

    monkeypatch.setattr(
        "dfcx_scrapi.core.intents.Intents.__init__", mock_intents_init
    )

    monkeypatch.setattr(
        "dfcx_scrapi.core.flows.Flows.__init__", mock_flows_init
    )

    monkeypatch.setattr(
        "dfcx_scrapi.core.pages.Pages.__init__", mock_pages_init
    )

    monkeypatch.setattr(
        "dfcx_scrapi.core.transition_route_groups.TransitionRouteGroups.__init__",
        mock_route_groups_init,
    )

    yield mock_credentials_from_file


# Test init with creds_path
def test_dffx_init_creds_path(mock_dffx_setup, test_config):
    mock_creds = mock_dffx_setup
    dffx = DataframeFunctions(creds_path=test_config["creds_path"])

    assert dffx.creds == test_config["creds_object"]
    mock_creds.assert_called_once_with(test_config["creds_path"])


# Test init with creds_dict
def test_dffx_init_creds_dict(mock_dffx_setup, test_config):
    mock_creds = mock_dffx_setup
    dffx = DataframeFunctions(creds_path=test_config["creds_dict"])

    assert dffx.creds == test_config["creds_object"]
    mock_creds.assert_called_once_with(test_config["creds_dict"])


# Test init with creds object
def test_dffx_init_creds_object(mock_dffx_setup, test_config):
    dffx = DataframeFunctions(creds=test_config["creds_object"])

    assert dffx.creds == test_config["creds_object"]



""" Compare the fullfillments of en vs french langauge or etc"""
import pandas as pd

from dfcx_scrapi.core.flows import Flows

from .fullfillment_helper import (
    PagesChild,
    get_entry_ff,
    get_param_ff,
    get_route_ff,
)


def en_vs_lang(agent_id,flows,lang):
    """Compares fulfillment coverage between English and a specified language.

    This function analyzes the fulfillment configurations (entry fulfillments,
    parameter fulfillments, and route fulfillments) for a given Dialogflow CX
    agent and a set of flows. It compares the coverage of the specified language
    (`lang`) with the English language ('en'), generating dataframes that
    highlight any discrepancies in fulfillment setup.

    Args:
        agent_id: The ID of the Dialogflow CX agent.
        flows: A list of flow display names to analyze.
        lang: The language code to compare against English (e.g., 'fr-ca').

    Returns:
        A tuple containing:
        - entry_df: DataFrame with statistics on entry fulfillment coverage.
        - param_df: DataFrame with statistics on parameter fulfillment coverage.
        - route_df: DataFrame with statistics on route fulfillment coverage.
        - result: A boolean indicating if all elements have agent responses
        in the specified language.
    """
    entry_columns = ['flow','page', 'text_entry_en', f'text_entry_{lang}',
                     'payload_entry_en', f'payload_entry_{lang}']
    entry_df = pd.DataFrame(columns=entry_columns)
    params_columns =['flow','page','parameter','text_param_en',
                     f'text_param_{lang}','payload_param_en',
                     f'payload_param_{lang}']
    param_df = pd.DataFrame(columns=params_columns)
    route_columns=['flow','page','route','text_route_en',
                   f'text_route_{lang}', 'payload_route_en',
                   f'payload_route_{lang}']
    route_df = pd.DataFrame(columns=route_columns)
    flowobj=Flows()
    pagesobj=PagesChild()
    for flow in flows:
        flow_details=flowobj.get_flow_by_display_name(display_name=flow,
                                               agent_id=agent_id)
        flow_id=flow_details.name
        pages_list=pagesobj.list_pages(flow_id=flow_id)

        for page in pages_list:
            page_name=page.display_name
            p_entry_en=0
            t_entry_en=0
            #getting entry fullfillment details
            p_entry_en,t_entry_en=get_entry_ff(page=page,language_code='en')

            if p_entry_en >0 or t_entry_en >0:
                p_entry_lang,t_entry_lang=get_entry_ff(
                    page_id=page.name,
                    language_code=lang)
                new_row = pd.DataFrame({
                    'flow': [flow],
                    'page': [page_name],
                    'text_entry_en':[t_entry_en] ,
                    f'text_entry_{lang}': [t_entry_lang],
                    'payload_entry_en':[p_entry_en],
                    f'payload_entry_{lang}': [p_entry_lang]
                    })
                entry_df = pd.concat([entry_df, new_row], ignore_index=True)

            #getting fullfillemnt in Parameters
            for idx,param in enumerate(page.form.parameters):
                param_name=param.display_name
                p_param_en,t_param_en=get_param_ff(param=param,language_code='en')
                if p_param_en> 0 or t_param_en >0:
                    p_param_lang,t_param_lang=get_param_ff(page_id=page.name,
                                                           idx=idx,
                                                           language_code='fr-ca')

                    new_row = pd.DataFrame({
                        'flow': [flow],
                        'page': [page_name],
                        'parameter' : [param_name],
                        'text_param_en':[t_param_en] ,
                        f'text_param_{lang}': [t_param_lang],
                        'payload_param_en':[p_param_en],
                        f'payload_param_{lang}': [p_param_lang]
                        })
                    param_df = pd.concat([param_df, new_row], ignore_index=True)

            #getting fullfillment details in page routes
            for idx,route in enumerate(page.transition_routes):
                route_name=route.name
                p_route_en,t_route_en=get_route_ff(route=route,language_code='en')
                if p_route_en>0 or t_route_en>0:
                    p_route_lang,t_route_lang=get_route_ff(page_id=page.name,
                                                           idx=idx,
                                                           language_code='fr-ca')

                    new_row = pd.DataFrame({
                        'flow': [flow],
                        'page': [page_name],
                        'route' : [route_name],
                        'text_route_en':[t_route_en] ,
                        f'text_route_{lang}': [t_route_lang],
                        'payload_route_en':[p_route_en],
                        f'payload_route_{lang}': [p_route_lang]
                        })
                    route_df=pd.concat([route_df, new_row],
                                         ignore_index=True)
    condition1 = (
        (entry_df.iloc[:, 2] != entry_df.iloc[:, 3]) |
        (entry_df.iloc[:, 4] != entry_df.iloc[:, 5])
    )
    condition2 = (
        (param_df.iloc[:, 3] != param_df.iloc[:, 4]) |
        (param_df.iloc[:, 5] != param_df.iloc[:, 6])
    )
    condition3 =(
        (route_df.iloc[:, 3] != route_df.iloc[:, 4]) |
        (route_df.iloc[:, 5] != route_df.iloc[:, 6])
    )

    result1 = entry_df[condition1]
    result2 = param_df[condition2]
    result3 = route_df[condition3]
    if result1.empty and result2.empty and result3.empty:
        result=True
    else:
        result=False

    return entry_df,param_df,route_df,result



"""Tools class for Generative Agents."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from google.cloud.dialogflowcx_v3beta1 import services, types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import scrapi_base


class Tools(scrapi_base.ScrapiBase):
    """Core Class for Tools Resource methods."""
    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        agent_id: str = None,
        tool_id: str = None,
        tools_map: Dict[str, str] = None
    ):
        super().__init__(
            creds_path=creds_path, creds_dict=creds_dict,
            creds=creds, scope=scope
        )

        self.agent_id = agent_id
        self.tool_id = tool_id
        self.tools_map = tools_map

    @staticmethod
    def build_open_api_tool(
        display_name: str, spec: str, description: str = None):
        """Helper method to build OpenAPI tool specs."""
        return types.Tool(
            display_name=display_name,
            description=description,
            open_api_spec=types.Tool.OpenApiTool(text_schema=spec)
            )

    @scrapi_base.api_call_counter_decorator
    def get_tools_map(self, agent_id: str, reverse: bool = False):
        """Returns a map of tool names to tool IDs"""
        if reverse:
            tool_map = {
                tool.display_name: tool.name
                for tool in self.list_tools(agent_id)
            }

        else:
            tool_map = {
                tool.name: tool.display_name
                for tool in self.list_tools(agent_id)
            }

        return tool_map

    @scrapi_base.api_call_counter_decorator
    def list_tools(self, agent_id: str):
        """Returns a list of tools for a given agent"""
        request = types.tool.ListToolsRequest(parent=agent_id)
        client_options = self._set_region(agent_id)
        client = services.tools.ToolsClient(
            client_options=client_options, credentials=self.creds
        )
        response = client.list_tools(request=request)

        return list(response)

    @scrapi_base.api_call_counter_decorator
    def get_tool(self, tool_id: str):
        """Get the specified Tool ID."""
        request = types.tool.GetToolRequest(name=tool_id)
        client_options = self._set_region(tool_id)
        client = services.tools.ToolsClient(
            client_options=client_options, credentials=self.creds
        )

        return client.get_tool(request=request)

    @scrapi_base.api_call_counter_decorator
    def create_tool(self, agent_id: str, obj: types.Tool = None, **kwargs):
        """Create an Agent Tool."""

        request = types.tool.CreateToolRequest()

        if obj:
            tool_obj = obj
            tool_obj.name = ""
        else:
            tool_obj = types.Tool()

            # set optional args as tool attributes
            for key, value in kwargs.items():
                setattr(tool_obj, key, value)

        request.parent = agent_id
        request.tool = tool_obj

        client_options = self._set_region(agent_id)
        client = services.tools.ToolsClient(
            client_options=client_options, credentials=self.creds
        )

        response = client.create_tool(request)

        return response

    @scrapi_base.api_call_counter_decorator
    def update_tool(self, tool_id: str, obj: types.Tool = None, **kwargs):
        """Update a single Tool with the specific object or kwargs."""
        if obj:
            tool = obj
            tool.name = tool_id
        else:
            tool = self.get_tool(tool_id)

        # set optional tool attributes
        for key, value in kwargs.items():
            setattr(tool, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(tool_id)
        client = services.tools.ToolsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.update_tool(tool=tool, update_mask=mask)

        return response

    @scrapi_base.api_call_counter_decorator
    def delete_tool(self, tool_id: str = None, obj: types.Tool = None):
        """Deletes a single Agent Tool resource."""
        if not tool_id:
            tool_id = self.tool_id

        if obj:
            tool_id = obj.name

        client_options = self._set_region(tool_id)
        client = services.tools.ToolsClient(
            client_options=client_options, credentials=self.creds
        )

        client.delete_tool(name=tool_id)
